"""Strategies do the merging between the files from template and the target.

Strategies are specified in the ``scaraplate.yaml`` file located in the root
of the template dir.

Sample ``scaraplate.yaml`` excerpt:

::

    default_strategy: scaraplate.strategies.Overwrite
    strategies_mapping:
      setup.py: scaraplate.strategies.TemplateHash
      src/*/__init__.py: scaraplate.strategies.IfMissing
      package.json:
        strategy: mypackage.mymodule.MyPackageJson
        config:
          my_key: True
      "src/{{ cookiecutter.myvariable }}.md": scaraplate.strategies.IfMissing


The strategy should be an importable Python class which implements
:class:`.Strategy`.

``default_strategy`` and ``strategies_mapping`` keys are the required ones.

The strategy value might be either a string (specifying a Python class),
or a dict of two keys -- ``strategy`` and ``config``. The first form
is just a shortcut for specifying a strategy with an empty config.

``config`` would be passed to the Strategy's ``__init__`` which would
be validated with the inner ``Schema`` class.
"""
import abc
import hashlib
import io
import re
from collections import OrderedDict
from configparser import ConfigParser
from typing import Any, BinaryIO, Dict, Iterable, List, Optional, Sequence, Set, Tuple

from marshmallow import Schema, ValidationError, fields, validates_schema
from marshmallow.validate import Range
from packaging.requirements import Requirement

from . import fields as scaraplate_fields
from .compat import marshmallow_load_data, marshmallow_pass_original_for_many
from .template import TemplateMeta

Pattern = Any  # re.Pattern since py3.7


def detect_newline(
    *file_contents_args: Optional[BinaryIO], default: bytes = b"\n"
) -> bytes:
    for file_contents in file_contents_args:
        if file_contents is None:
            continue
        assert file_contents.tell() == 0
        line_with_newline = file_contents.readline()
        file_contents.seek(0)

        line_without_newline = line_with_newline.rstrip(b"\r\n")
        if len(line_with_newline) != len(line_without_newline):
            # There's a newline at the end of the line -- use it
            newline = line_with_newline[len(line_without_newline) :]
            return newline

    return default


class NoExtraKeysSchema(Schema):
    """Marshmallow schema which raises an error for unknown keys
    in the input.

    Supposedly this won't be needed with marshmallow 3, as this feature
    has been built in,
    see https://marshmallow.readthedocs.io/en/3.0/quickstart.html#handling-unknown-fields  # noqa
    """

    @validates_schema(pass_original=True)
    def check_unknown_fields(self, data, original_data, **kwargs):
        # https://marshmallow.readthedocs.io/en/stable/extending.html#validating-original-input-data  # noqa
        original_data = marshmallow_pass_original_for_many(original_data, self.many)
        for item in original_data:
            unknown = set(item.keys()) - set(self.fields)
            if unknown:
                raise ValidationError({field: "Unknown field" for field in unknown})


class ConfigKeySchema(NoExtraKeysSchema):
    """A key in an INI-like config file."""

    sections = scaraplate_fields.Pattern(required=True)
    keys = scaraplate_fields.Pattern(required=True)


class ConfigSectionSchema(NoExtraKeysSchema):
    """A key in an INI-like config file."""

    sections = scaraplate_fields.Pattern(required=True)
    excluded_keys = scaraplate_fields.Pattern(missing=None)


class Strategy(abc.ABC):
    """The abstract base class for a scaraplate Strategy.

    To implement and use a custom strategy, the following needs to be done:

    1. Create a new Python class which implements :class:`.Strategy`
    2. Override the inner ``Schema`` class if you need
       to configure your strategy from ``scaraplate.yaml``.
    3. Implement the ``apply`` method.

    Assuming that the new strategy class is importable in the Python
    environment in which scaraplate is run, to use the strategy
    you need to specify it in ``scaraplate.yaml``, e.g.

    ::

        strategies_mapping:
          myfile.txt: mypackage.mymodule.MyStrategy

    """

    def __init__(
        self,
        *,
        target_contents: Optional[BinaryIO],
        template_contents: BinaryIO,
        template_meta: TemplateMeta,
        config: Dict[str, Any],
    ) -> None:
        """Init the strategy.

        :param target_contents: The file contents in the target project.
            ``None`` if the file doesn't exist.
        :param template_contents: The file contents from the template
            (after cookiecutter is applied).
        :param template_meta: Template metadata,
            see :class:`scaraplate.template.TemplateMeta`.
        :param config: The strategy config from ``scaraplate.yaml``.
            It is validated in this ``__init__`` with the inner
            ``Schema`` class.
        """
        self.target_contents = target_contents
        self.template_contents = template_contents
        self.template_meta = template_meta
        self.config = marshmallow_load_data(self.Schema, config)

    @abc.abstractmethod
    def apply(self) -> BinaryIO:
        """Apply the Strategy.

        :return: The resulting file contents which would overwrite
            the target file.
        """
        pass

    class Schema(NoExtraKeysSchema):
        """An empty default schema which doesn't accept any parameters."""


class Overwrite(Strategy):
    """A simple strategy which always overwrites the target files
    with the ones from the template.
    """

    def apply(self) -> BinaryIO:
        return self.template_contents


class IfMissing(Strategy):
    """A strategy which writes the file from the template only
    if it doesn't exist in target.
    """

    def apply(self) -> BinaryIO:
        if self.target_contents is None:
            return self.template_contents
        else:
            return self.target_contents


class SortedUniqueLines(Strategy):
    """A strategy which combines both template and target files,
    sorts the combined lines and keeps only unique ones.

    However, the comments in the beginning of the files are treated
    differently. They would be stripped from the target and replaced
    with the ones from the template. The most common usecase for this
    are the License headers.

    Sample ``scaraplate.yaml`` excerpt:

    ::

        strategies_mapping:
          MANIFEST.in:
            strategy: scaraplate.strategies.SortedUniqueLines
          .gitignore:
            strategy: scaraplate.strategies.SortedUniqueLines
    """

    def apply(self) -> BinaryIO:
        newline = detect_newline(self.target_contents, self.template_contents)

        header_lines, out_lines = self.split_header(
            self.template_contents.read().decode().splitlines()
        )
        if self.target_contents is not None:
            # Header from the target is ignored: it will be overridden
            # with template.
            _, target_lines = self.split_header(
                self.target_contents.read().decode().splitlines()
            )
            out_lines.extend(target_lines)

        # Keep unique lines and sort them.
        #
        # Note that `set` is not guaranteed to preserve the original
        # order, so we need to compare by both casefolded str and
        # the original to ensure the stable order for the same strings
        # written in different cases.
        sorted_lines = sorted(set(out_lines), key=lambda s: (s.casefold(), s))

        sorted_lines = [line for line in sorted_lines if line]
        sorted_lines.append("")  # trailing newline

        return io.BytesIO(newline.join(map(str.encode, header_lines + sorted_lines)))

    def split_header(self, lines: Sequence[str]) -> Tuple[List[str], List[str]]:
        comment_pattern = self.config["comment_pattern"]
        it = iter(lines)

        header_lines = []
        to_sort_lines = []
        for line in it:
            if comment_pattern.match(line) is not None or not line.strip():
                header_lines.append(line)
            else:
                to_sort_lines.append(line)
                break
        to_sort_lines.extend(it)

        return header_lines, to_sort_lines

    class Schema(NoExtraKeysSchema):
        r"""Allowed params:

        - ``comment_pattern`` [``^ *([;#%]|//)``] -- a PCRE pattern which should
          match the line with a comment.
        """

        comment_pattern = scaraplate_fields.Pattern(
            missing=re.compile(r"^ *([;#%]|//)")
        )


class TemplateHash(Strategy):
    """A strategy which appends to the target file a git commit hash of
    the template being applied; and the subsequent applications of
    the same template for this file are ignored until the HEAD commit
    of the template changes.

    This strategy is useful when a file needs to be different from
    the template but there's no suitable automated strategy yet,
    so it should be manually resynced on template updates.

    This strategy overwrites the target file on each new commit
    in the template. There's also a :class:`.RenderedTemplateFileHash`
    strategy which does it less frequently: only when the source file
    from the template has changes.

    Sample ``scaraplate.yaml`` excerpt:

    ::

        strategies_mapping:
          setup.py:
            strategy: scaraplate.strategies.TemplateHash
            config:
              line_comment_start: '#'
              max_line_length: 87
              max_line_linter_ignore_mark: '  # noqa'
          Jenkinsfile:
            strategy: scaraplate.strategies.TemplateHash
            config:
              line_comment_start: '//'

    This would result in the following:

    setup.py:

    ::

        ...file contents...

        # Generated by https://github.com/rambler-digital-solutions/scaraplate
        # From https://github.com/rambler-digital-solutions/scaraplate-example-template/commit/1111111111111111111111111111111111111111  # noqa

    Jenkinsfile:

    ::

        ...file contents...

        // Generated by https://github.com/rambler-digital-solutions/scaraplate
        // From https://github.com/rambler-digital-solutions/scaraplate-example-template/commit/1111111111111111111111111111111111111111

    """

    def searched_comment_contents(self) -> List[str]:
        return self.comment_contents()

    def comment_contents(self) -> List[str]:
        line_comment_start = self.config["line_comment_start"]

        comment_lines = [
            "Generated by https://github.com/rambler-digital-solutions/scaraplate"
        ]
        if self.template_meta.is_git_dirty:
            comment_lines.append(f"From (dirty) {self.template_meta.commit_url}")
        else:
            comment_lines.append(f"From {self.template_meta.commit_url}")

        return [f"{line_comment_start} {line}" for line in comment_lines]

    def render_comment(self, comment_lines: Sequence[str], *, newline: bytes) -> bytes:
        comment_lines = [self._maybe_add_linter_ignore(line) for line in comment_lines]
        return b"".join(line.encode("ascii") + newline for line in comment_lines)

    def _maybe_add_linter_ignore(self, line: str) -> str:
        line_length = self.config["max_line_length"]
        ignore_mark = self.config["max_line_linter_ignore_mark"]
        if line_length and len(line) >= line_length:
            return f"{line}{ignore_mark}"
        return line

    def apply(self) -> BinaryIO:
        newline = detect_newline(self.target_contents, self.template_contents)

        searched_comment = self.render_comment(
            self.searched_comment_contents(), newline=newline
        )
        appended_comment = self.render_comment(self.comment_contents(), newline=newline)
        if self.target_contents is not None:
            target_text = self.target_contents.read()
            if searched_comment in target_text and not self.template_meta.is_git_dirty:
                # Hash hasn't changed -- keep the target.
                self.target_contents.seek(0)
                return self.target_contents

        # Convert template newlines to the target file newlines:
        template_lines = self.template_contents.read().splitlines()
        out_bytes = b"".join(line + newline for line in template_lines)

        out_bytes += newline + appended_comment
        return io.BytesIO(out_bytes)

    class Schema(NoExtraKeysSchema):
        """Allowed params:

        - ``line_comment_start`` [``#``] -- The prefix which should be used
          to start a line comment.
        - ``max_line_length`` [`None`] -- The maximum line length for
          the appended line comments after which
          the ``max_line_linter_ignore_mark`` suffix should be added.
        - ``max_line_linter_ignore_mark`` [``# noqa``] -- The linter's
          line ignore mark for the appended line comments which are
          longer than ``max_line_length`` columns. The default ``# noqa``
          mark silences ``flake8``.
        """

        line_comment_start = fields.String(missing="#")
        max_line_length = fields.Int(missing=None, validate=Range(min=10))
        max_line_linter_ignore_mark = fields.String(missing="  # noqa")


class RenderedTemplateFileHash(TemplateHash):
    """A strategy which appends to the target file a hash of
    the rendered template file; and the subsequent applications of
    the same template for this file are ignored until the rendered
    file has changes.

    This strategy is similar to :class:`.TemplateHash` with the difference
    that the target file is rewritten less frequently: only when
    the hash of the source file from the template is changed.

    .. versionadded:: 0.2

    Sample ``scaraplate.yaml`` excerpt:

    ::

        strategies_mapping:
            setup.py:
              strategy: scaraplate.strategies.RenderedTemplateFileHash
              config:
                line_comment_start: '#'
                max_line_length: 87
                max_line_linter_ignore_mark: '  # noqa'
            Jenkinsfile:
              strategy: scaraplate.strategies.RenderedTemplateFileHash
              config:
                line_comment_start: '//'

    This would result in the following:

    setup.py:

    ::

        ...file contents...

        # Generated by https://github.com/rambler-digital-solutions/scaraplate
        # RenderedTemplateFileHash d2671228e3dfc3e663bfaf9b5b151ce8
        # From https://github.com/rambler-digital-solutions/scaraplate-example-template/commit/1111111111111111111111111111111111111111  # noqa

    Jenkinsfile:

    ::

        ...file contents...

        // Generated by https://github.com/rambler-digital-solutions/scaraplate
        // RenderedTemplateFileHash d2as1228eb7233e663bfaf9b5b151ce8
        // From https://github.com/rambler-digital-solutions/scaraplate-example-template/commit/1111111111111111111111111111111111111111

    """

    def searched_comment_contents(self) -> List[str]:
        line_comment_start = self.config["line_comment_start"]

        rendered_template_file_hash = hashlib.md5(
            self.template_contents.read()
        ).hexdigest()
        self.template_contents.seek(0)
        comment_lines = [
            "Generated by https://github.com/rambler-digital-solutions/scaraplate",
            f"RenderedTemplateFileHash {rendered_template_file_hash}",
        ]

        return [f"{line_comment_start} {line}" for line in comment_lines]

    def comment_contents(self) -> List[str]:
        line_comment_start = self.config["line_comment_start"]

        comment_lines = []
        if self.template_meta.is_git_dirty:
            comment_lines.append(f"From (dirty) {self.template_meta.commit_url}")
        else:
            comment_lines.append(f"From {self.template_meta.commit_url}")

        return self.searched_comment_contents() + [
            f"{line_comment_start} {line}" for line in comment_lines
        ]


class ConfigParserMerge(Strategy):
    """A strategy which merges INI-like files (with :mod:`configparser`).

    The resulting file is the one from the template with
    the following modifications:

    - Comments are stripped
    - INI file is reformatted (whitespaces are cleaned, sections
      and keys are sorted)
    - Sections specified in the ``preserve_sections`` config list are
      preserved from the target file.
    - Keys specified in the ``preserve_keys`` config list are
      preserved from the target file.

    This strategy cannot be used to merge config files which contain
    keys without a preceding section declaration
    (e.g. ``.editorconfig`` won't work).

    Sample ``scaraplate.yaml`` excerpt:

    ::

        strategies_mapping:
          .pylintrc:
            strategy: scaraplate.strategies.ConfigParserMerge
            config:
              preserve_sections: []
              preserve_keys:
              - sections: ^MASTER$
                keys: ^extension-pkg-whitelist$
              - sections: ^TYPECHECK$
                keys: ^ignored-

          tox.ini:
            strategy: scaraplate.strategies.ConfigParserMerge
            config:
              preserve_sections:
              - sections: ^tox$
              preserve_keys:
              - sections: ^testenv
                keys: ^extras$

          pytest.ini:
            strategy: scaraplate.strategies.ConfigParserMerge
            config:
              preserve_sections: []
              preserve_keys:
              - sections: ^pytest$
                keys: ^python_files$

          .isort.cfg:
            strategy: scaraplate.strategies.ConfigParserMerge
            config:
              preserve_sections: []
              preserve_keys:
              - sections: ^settings$
                keys: ^known_third_party$
    """

    def apply(self) -> BinaryIO:
        newline = detect_newline(self.target_contents, self.template_contents)

        template_parser = self.parse_config(self.template_contents, source="<template>")

        target_parser: Optional[ConfigParser] = None

        if self.target_contents is not None:
            target_parser = self.parse_config(self.target_contents, source="<target>")

        self.merge_configs(template_parser, target_parser, newline=newline)

        return self.parser_to_pretty_output(template_parser, newline=newline)

    def parse_config(self, data: BinaryIO, source: str) -> ConfigParser:
        # We don't need to treat the `[DEFAULT]` section as actually
        # a one which provides defaults to other sections, so we disable
        # this by mocking the `default_section`.
        parser = ConfigParser(
            default_section="__scaraplate_internal_nonexisting_default"
        )

        text = data.read().decode()
        # Configparser doesn't like \r\n and \r newlines, so let's replace
        # them explicitly:
        text = "".join(f"{line}\n" for line in text.splitlines())
        parser.read_string(text, source=source)
        return parser

    def merge_configs(
        self,
        template_parser: ConfigParser,
        target_parser: Optional[ConfigParser],
        *,
        newline: bytes,
    ) -> None:
        if target_parser is None:
            return

        for ini_key in self.config["preserve_sections"]:
            self.maybe_preserve_sections(
                template_parser,
                target_parser,
                ini_key["sections"],
                ini_key["excluded_keys"],
            )

        for ini_key in self.config["preserve_keys"]:
            self.maybe_preserve_key(
                template_parser, target_parser, ini_key["sections"], ini_key["keys"]
            )

    def maybe_preserve_key(
        self,
        template_parser: ConfigParser,
        target_parser: ConfigParser,
        sections_pattern: Pattern,
        keys_pattern: Pattern,
    ) -> None:
        for section in target_parser.sections():
            if sections_pattern.match(section):
                for key, value in target_parser[section].items():
                    if keys_pattern.match(key):
                        self.ensure_section(template_parser, section)
                        template_parser[section][key] = value

    def maybe_preserve_sections(
        self,
        template_parser: ConfigParser,
        target_parser: ConfigParser,
        sections_pattern: Pattern,
        ignore_keys_pattern: Optional[Pattern],
    ) -> None:
        for section in target_parser.sections():
            if sections_pattern.match(section):
                section_data = dict(target_parser[section])

                if ignore_keys_pattern is not None:
                    for key, value in template_parser[section].items():
                        if ignore_keys_pattern.match(key):
                            section_data[key] = value

                template_parser[section] = section_data

    def ensure_section(self, parser: ConfigParser, section: str) -> None:
        if not parser.has_section(section):
            parser.add_section(section)

    def parser_to_pretty_output(
        self, parser: ConfigParser, *, newline: bytes
    ) -> BinaryIO:
        parser = self._sorted_configparser(parser)

        content = self._parser_to_str(parser).replace("\t", " " * 4)

        acc = []
        for line in content.splitlines():
            acc.append(line.rstrip())
        new_contents = newline.join(map(str.encode, acc))
        return io.BytesIO(new_contents)

    def _sorted_configparser(self, parser: ConfigParser) -> ConfigParser:
        out = ConfigParser(dict_type=OrderedDict)

        # `out.read_dict(parser)` might not work here as expected: the keys from
        # the `[DEFAULT]` section would be set in *all* sections instead of
        # just the `[DEFAULT]` one.
        out.read_string(self._parser_to_str(parser))

        for section in out._sections:  # type: ignore
            section_ = OrderedDict(
                sorted(out._sections[section].items())  # type: ignore
            )
            out._sections[section] = section_  # type: ignore

        out._sections = OrderedDict(sorted(out._sections.items()))  # type: ignore
        return out

    def _parser_to_str(self, parser: ConfigParser) -> str:
        buffer = io.StringIO()
        parser.write(buffer)
        return buffer.getvalue()

    class Schema(NoExtraKeysSchema):
        """Allowed params:

        - ``preserve_keys`` (required) -- the list of config keys
          which should be preserved from the target file. Values schema:

            + ``sections`` (required) -- a PCRE pattern matching
              sections containing the keys to preserve.
            + ``keys`` (required) -- a PCRE pattern matching keys
              in the matched sections.

        - ``preserve_sections`` (required) -- the list of config sections
          which should be preserved from the target file. If the matching
          section exists in the template, it would be fully overwritten.
          Values schema:

            + ``sections`` (required) -- a PCRE pattern matching
              sections which should be preserved from the target.
            + ``excluded_keys`` [`None`] -- a PCRE pattern matching
              the keys which should not be overwritten in the template
              when preserving the section.

        """

        preserve_keys = fields.Nested(ConfigKeySchema, many=True, required=True)
        preserve_sections = fields.Nested(ConfigSectionSchema, many=True, required=True)


class SetupCfgMerge(ConfigParserMerge):
    r"""A strategy which merges the Python's ``setup.cfg`` file.

    Based on the :class:`.ConfigParserMerge` strategy, additionally
    containing a ``merge_requirements`` config option for merging
    the lists of Python requirements between the files.

    Sample ``scaraplate.yaml`` excerpt:

    ::

        strategies_mapping:
          setup.cfg:
            strategy: scaraplate.strategies.SetupCfgMerge
            config:
              merge_requirements:
              - sections: ^options$
                keys: ^install_requires$
              - sections: ^options\.extras_require$
                keys: ^develop$
              preserve_keys:
              - sections: ^tool:pytest$
                keys: ^testpaths$
              - sections: ^build$
                keys: ^executable$
              preserve_sections:
              - sections: ^mypy-
              - sections: ^options\.data_files$
              - sections: ^options\.entry_points$
              - sections: ^options\.extras_require$
    """

    class Schema(ConfigParserMerge.Schema):
        __doc__ = (
            ConfigParserMerge.Schema.__doc__  # type: ignore
            + """

        - ``merge_requirements`` (required) -- the list of config
          keys containing the lists of Python requirements which should
          be merged together. Values schema:

            + ``sections`` (required) -- a PCRE pattern matching
              sections containing the keys with requirements.
            + ``keys`` (required) -- a PCRE pattern matching keys
              in the matched sections.
        """
        )

        merge_requirements = fields.Nested(ConfigKeySchema, many=True, required=True)

    def merge_configs(
        self,
        template_parser: ConfigParser,
        target_parser: Optional[ConfigParser],
        *,
        newline: bytes,
    ) -> None:
        keys: Set[Tuple[str, str]] = set()

        for ini_key in self.config["merge_requirements"]:
            if target_parser is not None:
                for section in target_parser.sections():
                    if ini_key["sections"].match(section):
                        for key, value in target_parser[section].items():
                            if ini_key["keys"].match(key):
                                keys.add((section, key))

            for section in template_parser.sections():
                if ini_key["sections"].match(section):
                    for key, value in template_parser[section].items():
                        if ini_key["keys"].match(key):
                            keys.add((section, key))

        for section, key in keys:
            self._merge_requirements(
                template_parser, target_parser, section, key, newline=newline
            )

        super().merge_configs(template_parser, target_parser, newline=newline)

    def _merge_requirements(
        self,
        template_parser: ConfigParser,
        target_parser: Optional[ConfigParser],
        section: str,
        key: str,
        *,
        newline: bytes,
    ) -> None:
        "Merge in the requirements from template to the target."

        template_requirements = self._parse_requirements(template_parser, section, key)
        if target_parser is not None:
            target_requirements = self._parse_requirements(target_parser, section, key)
        else:
            target_requirements = []

        def normalize_requirement(requirement):
            return self._requirement_name(requirement).lower()

        existing_requirement_names = set(
            map(normalize_requirement, target_requirements)
        )
        wanted_requirements = target_requirements

        for requirement in template_requirements:
            name = normalize_requirement(requirement)
            if name not in existing_requirement_names:
                wanted_requirements.append(requirement)

        wanted_requirements = sorted(wanted_requirements, key=str.casefold)

        result = self._dump_setupcfg_requirements(wanted_requirements, newline=newline)

        self.ensure_section(template_parser, section)
        template_parser[section][key] = result

        if target_parser is not None:
            self.ensure_section(target_parser, section)
            target_parser[section][key] = result

    def _parse_requirements(
        self, parser: ConfigParser, section: str, key: str
    ) -> List[str]:
        try:
            requirements = parser[section][key]
        except KeyError:
            return []

        return self._parse_setupcfg_requirements(requirements)

    def _parse_setupcfg_requirements(self, requirements: str) -> List[str]:
        return [r for r in map(str.strip, requirements.split()) if r]

    def _requirement_name(self, full_requirement: str) -> str:
        requirement = Requirement(full_requirement)
        return requirement.name

    def _dump_setupcfg_requirements(
        self, requirements: Iterable[str], *, newline: bytes
    ) -> str:
        # Leave first element empty to produce nicer cfg lists like:
        #   install_requires =
        #       foo==1.0.0
        #       bar==2.0.0
        # instead of
        #   install_requires = foo==1.0.0
        #       bar==2.0.0
        acc = [""] + [item for item in requirements]
        return newline.decode().join(acc)



import collections.abc
import importlib
from pathlib import Path
from typing import Any, Dict, Mapping, NamedTuple, Optional, Type, Union, cast

import jinja2
import yaml

from .cookiecutter import CookieCutterContext, CookieCutterContextDict, ScaraplateConf
from .gitremotes import GitRemote
from .strategies import Strategy

env = jinja2.Environment(undefined=jinja2.StrictUndefined)


class StrategyNode(NamedTuple):
    strategy: Type[Strategy]
    config: Dict[str, Any]


class ScaraplateYamlOptions(NamedTuple):
    git_remote_type: Optional[Type[GitRemote]]
    cookiecutter_context_type: Type[CookieCutterContext]


class ScaraplateYamlStrategies(NamedTuple):
    default_strategy: StrategyNode
    strategies_mapping: Mapping[str, StrategyNode]


def get_scaraplate_yaml_options(template_path: Path) -> ScaraplateYamlOptions:
    config = yaml.safe_load((template_path / "scaraplate.yaml").read_text())

    git_remote_type_name = config.get("git_remote_type")
    git_remote_type = (
        class_from_str(git_remote_type_name, ensure_subclass=GitRemote)
        if git_remote_type_name is not None
        else None
    )
    assert git_remote_type is None or issubclass(git_remote_type, GitRemote)  # mypy

    cookiecutter_context_type_name = config.get("cookiecutter_context_type")
    cookiecutter_context_type = (
        class_from_str(
            cookiecutter_context_type_name, ensure_subclass=CookieCutterContext
        )
        if cookiecutter_context_type_name is not None
        else cast(Type[CookieCutterContext], ScaraplateConf)  # mypy
    )
    assert cookiecutter_context_type is None or issubclass(
        cookiecutter_context_type, CookieCutterContext
    )  # mypy

    return ScaraplateYamlOptions(
        git_remote_type=git_remote_type,
        cookiecutter_context_type=cookiecutter_context_type,
    )


def get_scaraplate_yaml_strategies(
    template_path: Path, cookiecutter_context_dict: CookieCutterContextDict
) -> ScaraplateYamlStrategies:
    config = yaml.safe_load((template_path / "scaraplate.yaml").read_text())
    default_strategy = _parse_strategy_node(
        "default_strategy", config["default_strategy"]
    )

    strategies_mapping: Dict[str, StrategyNode] = {
        env.from_string(path).render(
            cookiecutter=cookiecutter_context_dict
        ): _parse_strategy_node(str(path), strategy_node)
        for path, strategy_node in config["strategies_mapping"].items()
    }

    return ScaraplateYamlStrategies(
        default_strategy=default_strategy, strategies_mapping=strategies_mapping
    )


def _parse_strategy_node(path: str, raw: Union[str, Dict[str, Any]]) -> StrategyNode:
    if isinstance(raw, str):
        strategy = raw
        config: Dict[str, Any] = {}
    elif isinstance(raw, collections.abc.Mapping):
        strategy_ = raw.get("strategy")
        if not isinstance(strategy_, str):
            raise ValueError(
                f"Unexpected `strategy` value for {path}: "
                f"a string is expected, got {strategy_!r}"
            )
        strategy = strategy_  # mypy

        config = raw.get("config", {})
        if not isinstance(config, collections.abc.Mapping):
            raise ValueError(
                f"Unexpected strategy `config` value for {path}: "
                f"a mapping is expected, got {config!r}"
            )
    else:
        raise ValueError(f"Unexpected strategy value type for {path}: got {raw!r}")

    cls = class_from_str(strategy, ensure_subclass=Strategy)
    assert issubclass(cls, Strategy)  # mypy
    return StrategyNode(strategy=cls, config=config)


def class_from_str(ref: str, *, ensure_subclass: Optional[Type] = None) -> Type[object]:
    if not isinstance(ref, str) or "." not in ref:
        raise ValueError(
            f"A Python class reference must look like "
            f"`mypackage.mymodule.MyClass`, got {ref!r}"
        )
    module_s, cls_s = ref.rsplit(".", 1)
    module = importlib.import_module(module_s)
    cls = getattr(module, cls_s)
    if ensure_subclass is not None and (
        not issubclass(cls, ensure_subclass) or cls is ensure_subclass
    ):
        raise ValueError(f"`{cls}` is not a subclass of {ensure_subclass}")
    return cls



import contextlib
import json
from pathlib import Path
from unittest.mock import ANY

import pytest

from scaraplate.automation.base import automatic_rollup
from scaraplate.automation.git import (
    GitCloneProjectVCS,
    GitCloneTemplateVCS,
    strip_credentials_from_git_remote,
)


@contextlib.contextmanager
def git_unbare_repo(git_path: Path, call_git):
    call_git("git config --bool core.bare false", cwd=git_path)
    try:
        yield
    finally:
        call_git("git config --bool core.bare true", cwd=git_path)


def convert_bare_to_monorepo(dest: str, git_path: Path, call_git):
    with git_unbare_repo(git_path, call_git):
        dest_path = git_path / dest
        dest_path.mkdir()

        for child in git_path.iterdir():
            if child.name in (".git", dest):
                continue
            child.rename(dest_path / child.name)

        call_git("git add --all .", cwd=git_path)
        call_git('git commit -m "convert to monorepo"', cwd=git_path)
        call_git("git branch -f master2 master", cwd=git_path)


@pytest.mark.parametrize(
    "target_branch, clone_ref",
    [
        ("updates", None),
        ("updates", "master2"),
        ("updates", "master"),
        ("master", None),
        # master-master2 is excluded, because cloning a non-default branch and
        # committing the changes back to default is ridiculous
        ("master", "master"),
    ],
)
@pytest.mark.parametrize("monorepo_inner_path", ["inner", None])
def test_automatic_rollup(
    template_bare_git_repo: Path,
    project_bare_git_repo: Path,
    target_branch,
    clone_ref,
    monorepo_inner_path,
    call_git,
):
    if monorepo_inner_path is not None:
        convert_bare_to_monorepo(monorepo_inner_path, template_bare_git_repo, call_git)
        convert_bare_to_monorepo(monorepo_inner_path, project_bare_git_repo, call_git)

    assert "master" == call_git(
        "git rev-parse --abbrev-ref HEAD", cwd=project_bare_git_repo
    )
    master_commit_hash = call_git(
        "git rev-parse --verify master", cwd=project_bare_git_repo
    )
    template_commit_hash = call_git(
        f"git rev-parse --verify {clone_ref or 'master'}", cwd=template_bare_git_repo
    )

    automatic_rollup(
        template_vcs_ctx=GitCloneTemplateVCS.clone(
            clone_url=str(template_bare_git_repo),  # a clonable remote URL
            clone_ref=clone_ref,
            monorepo_inner_path=monorepo_inner_path,
        ),
        project_vcs_ctx=GitCloneProjectVCS.clone(
            clone_url=str(project_bare_git_repo),  # a clonable remote URL
            clone_ref=clone_ref,
            monorepo_inner_path=monorepo_inner_path,
            changes_branch=target_branch,
            commit_author="pytest <tests@none>",
        ),
        extra_context={"key1": "value1", "key2": "value2"},
    )

    target_branch_commit_hash = call_git(
        f"git rev-parse --verify {target_branch}", cwd=project_bare_git_repo
    )
    if target_branch == "master":
        # Ensure that master branch has advanced
        assert master_commit_hash != target_branch_commit_hash

    commit_message = call_git(
        f'git log --pretty=format:"%B" -n 1 {target_branch}', cwd=project_bare_git_repo
    )
    assert (clone_ref or "master") in commit_message  # template clone ref
    assert template_commit_hash in commit_message

    # Apply again (this time we expect no changes)
    automatic_rollup(
        template_vcs_ctx=GitCloneTemplateVCS.clone(
            clone_url=str(template_bare_git_repo),  # a clonable remote URL
            clone_ref=clone_ref,
            monorepo_inner_path=monorepo_inner_path,
        ),
        project_vcs_ctx=GitCloneProjectVCS.clone(
            clone_url=str(project_bare_git_repo),  # a clonable remote URL
            clone_ref=clone_ref,
            monorepo_inner_path=monorepo_inner_path,
            changes_branch=target_branch,
            commit_author="pytest <tests@none>",
        ),
        extra_context={"key1": "value1", "key2": "value2"},
    )

    target_branch_commit_hash_2 = call_git(
        f"git rev-parse --verify {target_branch}", cwd=project_bare_git_repo
    )
    assert target_branch_commit_hash == target_branch_commit_hash_2, call_git(
        f"git show -p {target_branch}", cwd=project_bare_git_repo
    )


def test_automatic_rollup_with_existing_target_branch(
    template_bare_git_repo: Path, project_bare_git_repo: Path, call_git
):
    target_branch = "update"

    assert "master" == call_git(
        "git rev-parse --abbrev-ref HEAD", cwd=project_bare_git_repo
    )
    master_commit_hash = call_git(
        "git rev-parse --verify master", cwd=project_bare_git_repo
    )

    # Create `update` branch with one version
    automatic_rollup(
        template_vcs_ctx=GitCloneTemplateVCS.clone(
            clone_url=str(template_bare_git_repo)  # a clonable remote URL
        ),
        project_vcs_ctx=GitCloneProjectVCS.clone(
            clone_url=str(project_bare_git_repo),  # a clonable remote URL
            changes_branch=target_branch,
            commit_author="pytest <tests@none>",
        ),
        extra_context={"key1": "first1", "key2": "first2"},
    )

    target_branch_commit_hash = call_git(
        f"git rev-parse --verify {target_branch}", cwd=project_bare_git_repo
    )

    # Apply again (but another version of template -- note the changed
    # extra_context
    automatic_rollup(
        template_vcs_ctx=GitCloneTemplateVCS.clone(
            clone_url=str(template_bare_git_repo)  # a clonable remote URL
        ),
        project_vcs_ctx=GitCloneProjectVCS.clone(
            clone_url=str(project_bare_git_repo),  # a clonable remote URL
            changes_branch=target_branch,
            commit_author="pytest <tests@none>",
        ),
        extra_context={"key1": "second1", "key2": "second2"},
    )

    master_commit_hash_2 = call_git(
        "git rev-parse --verify master", cwd=project_bare_git_repo
    )
    target_branch_commit_hash_2 = call_git(
        f"git rev-parse --verify {target_branch}", cwd=project_bare_git_repo
    )
    # The first push should be replaced with the second one:
    assert target_branch_commit_hash != target_branch_commit_hash_2
    # master should not be changed:
    assert master_commit_hash == master_commit_hash_2


@pytest.mark.parametrize("monorepo_inner_path", ["inner", None])
@pytest.mark.template_with_sense_vars
def test_automatic_rollup_preserves_template_dirname(
    template_bare_git_repo: Path,
    project_bare_git_repo: Path,
    monorepo_inner_path,
    call_git,
):
    target_branch = "update"

    if monorepo_inner_path is not None:
        convert_bare_to_monorepo(monorepo_inner_path, template_bare_git_repo, call_git)

    automatic_rollup(
        template_vcs_ctx=GitCloneTemplateVCS.clone(
            clone_url=str(template_bare_git_repo),  # a clonable remote URL
            monorepo_inner_path=monorepo_inner_path,
        ),
        project_vcs_ctx=GitCloneProjectVCS.clone(
            clone_url=str(project_bare_git_repo),  # a clonable remote URL
            changes_branch=target_branch,
            commit_author="pytest <tests@none>",
        ),
        extra_context={"key1": "value1", "key2": "value2"},
    )

    cookiecutter_context_text = call_git(
        f"git show {target_branch}:sense_vars", cwd=project_bare_git_repo
    )

    assert json.loads(cookiecutter_context_text) == {
        "_output_dir": ANY,
        "_template": monorepo_inner_path if monorepo_inner_path else "remote_template",
        "project_dest": "remote_project",
        "key1": "value1",
        "key2": "value2",
    }


@pytest.mark.parametrize(
    "remote_url, expected_url",
    [
        (
            "https://oauth2:xxxtokenxxx@mygitlab.com/myorg/myproj.git",
            "https://mygitlab.com/myorg/myproj.git",
        ),
        (
            "https://mygitlab.com/myorg/myproj.git",
            "https://mygitlab.com/myorg/myproj.git",
        ),
        # I suppose ssh urls cannot contain a password?
        ("git@mygitlab.com:myorg/myproj.git", "git@mygitlab.com:myorg/myproj.git"),
        ("/var/myrepos/mybare", "/var/myrepos/mybare"),
    ],
)
def test_strip_credentials_from_git_remote(remote_url, expected_url):
    assert expected_url == strip_credentials_from_git_remote(remote_url)



from contextlib import ExitStack
from pathlib import Path

import pytest

from scaraplate.cookiecutter import ScaraplateConf, SetupCfg


@pytest.mark.parametrize(
    "contents, expected_context",
    [
        (None, None),
        ("", {}),
        (
            """
[metadata]
name = holywarrior
""",
            {},
        ),
        (
            """
[tool:cookiecutter_context]
""",
            {},
        ),
        (
            """
[metadata]
name = holywarrior


[tool:cookiecutter_context]
metadata_author = Usermodel @ Rambler&Co
coverage_fail_under = 90
project_monorepo_name =
""",
            {
                "metadata_author": "Usermodel @ Rambler&Co",
                "coverage_fail_under": "90",
                "project_monorepo_name": "",
            },
        ),
    ],
)
def test_setupcfg(tempdir_path: Path, contents, expected_context):
    if contents is not None:
        (tempdir_path / "setup.cfg").write_text(contents)

    with ExitStack() as stack:
        if expected_context is None:
            stack.enter_context(pytest.raises(FileNotFoundError))

        cookiecutter_context = SetupCfg(tempdir_path)
        assert expected_context == cookiecutter_context.read()


@pytest.mark.parametrize(
    "contents, expected_context",
    [
        (None, None),
        ("", {}),
        (
            """
[foreignsection]
name = holywarrior
""",
            {},
        ),
        (
            """
[cookiecutter_context]
""",
            {},
        ),
        (
            """
[cookiecutter_context]
metadata_author = Usermodel @ Rambler&Co
coverage_fail_under = 90
project_monorepo_name =
""",
            {
                "metadata_author": "Usermodel @ Rambler&Co",
                "coverage_fail_under": "90",
                "project_monorepo_name": "",
            },
        ),
    ],
)
def test_scaraplate_conf(tempdir_path: Path, contents, expected_context):
    if contents is not None:
        (tempdir_path / ".scaraplate.conf").write_text(contents)

    with ExitStack() as stack:
        if expected_context is None:
            stack.enter_context(pytest.raises(FileNotFoundError))

        cookiecutter_context = ScaraplateConf(tempdir_path)
        assert expected_context == cookiecutter_context.read()



import os
import subprocess
import tempfile
from pathlib import Path

import pytest


@pytest.fixture
def tempdir_path():
    with tempfile.TemporaryDirectory() as tempdir_path:
        yield Path(tempdir_path).resolve()


@pytest.fixture
def init_git_and_commit(call_git):
    def _init_git_and_commit(path: Path, with_remote=True) -> None:
        call_git("git init", cwd=path)
        call_git("git add --all .", cwd=path)
        call_git('git commit -m "initial"', cwd=path)
        if with_remote:
            call_git(
                "git remote add origin https://gitlab.localhost/nonexisting/repo.git",
                cwd=path,
            )

    return _init_git_and_commit


@pytest.fixture
def call_git():
    def _call_git(shell_cmd: str, cwd: Path) -> str:
        env = {
            "GIT_AUTHOR_EMAIL": "pytest@scaraplate",
            "GIT_AUTHOR_NAME": "tests_scaraplate",
            "GIT_COMMITTER_EMAIL": "pytest@scaraplate",
            "GIT_COMMITTER_NAME": "tests_scaraplate",
            "GIT_CONFIG_NOSYSTEM": "1",
            "PATH": os.getenv("PATH", os.defpath),
        }
        out = subprocess.run(
            shell_cmd,
            shell=True,
            check=True,
            cwd=cwd,
            env=env,
            timeout=5,
            stdout=subprocess.PIPE,
        )
        stdout = out.stdout.decode().strip()
        print(stdout)
        return stdout

    return _call_git



import io
from contextlib import ExitStack
from typing import Any, BinaryIO, Dict, Optional, Type

import pytest
import yaml

from scaraplate import strategies
from scaraplate.config import class_from_str
from scaraplate.template import TemplateMeta


def pytest_collect_file(parent, file_path):
    # https://docs.pytest.org/en/latest/example/nonpython.html
    if file_path.suffix == ".yml" and file_path.name.startswith("test"):
        return YamlFile.from_parent(parent, path=file_path)


class YamlFile(pytest.File):
    def collect(self):
        spec = yaml.safe_load(self.path.open())
        strategy = spec["strategy"]
        testcases = spec["testcases"]

        for testcase in testcases:
            yield YamlItem.from_parent(
                self, name=testcase["name"], strategy=strategy, testcase=testcase
            )


class YamlItem(pytest.Item):
    def __init__(self, *, strategy, testcase, **kwargs):
        super().__init__(**kwargs)
        self.strategy = strategy
        self.testcase = testcase

    def runtest(self):
        strategy_type = getattr(strategies, self.strategy)
        run_strategy_test(
            strategy_type=strategy_type,
            template=self.testcase["template"],
            target=self.testcase["target"],
            is_git_dirty=self.testcase.get("is_git_dirty", False),
            out=self.testcase["out"],
            config=self.testcase.get("config", {}),
            raises=self.testcase.get("raises"),
        )

    def reportinfo(self):
        return self.fspath, 0, f"name: {self.name}"


def run_strategy_test(
    strategy_type: Type[strategies.Strategy],
    template: str,
    target: Optional[str],
    is_git_dirty: bool,
    out: Optional[str],
    config: Dict[str, Any],
    raises: Optional[str],
):
    if target is not None:
        target_contents: Optional[BinaryIO] = io.BytesIO(target.encode())
    else:
        target_contents = None

    template_contents = io.BytesIO(template.encode())

    raises_cls: Optional[Type] = class_from_str(raises) if raises is not None else None

    with ExitStack() as stack:
        if raises_cls is not None:
            assert out is None
            stack.enter_context(pytest.raises(raises_cls))

        strategy = strategy_type(
            target_contents=target_contents,
            template_contents=template_contents,
            template_meta=TemplateMeta(
                git_project_url=(
                    "https://github.com/rambler-digital-solutions/"
                    "scaraplate-example-template"
                ),
                commit_hash="1111111111111111111111111111111111111111",
                commit_url=(
                    "https://github.com/rambler-digital-solutions/"
                    "scaraplate-example-template"
                    "/commit/1111111111111111111111111111111111111111"
                ),
                is_git_dirty=is_git_dirty,
                head_ref="master",
            ),
            config=config,
        )

        strategy_out = strategy.apply().read().decode()
        assert out == strategy_out



import collections
from typing import Mapping

import click

from scaraplate.rollup import rollup as _rollup


def validate_extra_context(ctx, param, value):
    """Validate extra context."""
    # vendored from https://github.com/cookiecutter/cookiecutter/blob/673f773bfaf591b056d977c4ab82b45d90dce11e/cookiecutter/cli.py#L35-L46  # noqa
    for s in value:
        if "=" not in s:
            raise click.BadParameter(
                "EXTRA_CONTEXT should contain items of the form key=value; "
                "'{}' doesn't match that form".format(s)
            )

    # Convert tuple -- e.g.: (u'program_name=foobar', u'startsecs=66')
    # to dict -- e.g.: {'program_name': 'foobar', 'startsecs': '66'}
    return collections.OrderedDict(s.split("=", 1) for s in value) or None


@click.group()
@click.version_option()
def main():
    pass


@main.command()
@click.argument("template_dir", type=click.Path(exists=True, file_okay=False))
@click.argument("target_project_dir", type=click.Path(file_okay=False))
@click.argument("extra_context", nargs=-1, callback=validate_extra_context)
@click.option(
    "--no-input",
    is_flag=True,
    help=(
        "Do not prompt for missing data. Cookiecutter will use "
        "the defaults provided by the `cookiecutter.json` "
        "in the TEMPLATE_DIR."
    ),
)
def rollup(
    *,
    template_dir: str,
    target_project_dir: str,
    no_input: bool,
    extra_context: Mapping[str, str],
) -> None:
    """Rollup (apply) the cookiecutter template.

    The template from TEMPLATE_DIR is applied on top of TARGET_PROJECT_DIR.

    TEMPLATE_DIR must be a local path to the location of a git repo with
    the cookiecutter template to apply. It must contain `scaraplate.yaml`
    file in the root. The TEMPLATE_DIR must be in a git repo, because
    some strategies require the commit hash at HEAD and the git remote URL.

    TARGET_PROJECT_DIR should point to the directory where the template
    should be applied. Might not exist -- it will be created then.
    For monorepos this should point to the subproject inside the monorepo.

    EXTRA_CONTEXT is a list of `key=value` pairs, just like in
    `cookiecutter` command.
    """
    _rollup(
        template_dir=template_dir,
        target_project_dir=target_project_dir,
        no_input=no_input,
        extra_context=extra_context,
    )


if __name__ == "__main__":
    main(prog_name="scaraplate")



from pathlib import Path

import jinja2.exceptions
import pytest

import scaraplate.strategies
from scaraplate.config import (
    ScaraplateYamlOptions,
    ScaraplateYamlStrategies,
    StrategyNode,
    get_scaraplate_yaml_options,
    get_scaraplate_yaml_strategies,
)
from scaraplate.cookiecutter import CookieCutterContextDict, ScaraplateConf, SetupCfg
from scaraplate.gitremotes import GitHub


@pytest.mark.parametrize(
    "yaml_text, expected",
    [
        (
            """
git_remote_type: scaraplate.gitremotes.GitHub
""",
            ScaraplateYamlOptions(
                git_remote_type=GitHub, cookiecutter_context_type=ScaraplateConf
            ),
        ),
        (
            """
cookiecutter_context_type: scaraplate.cookiecutter.SetupCfg
""",
            ScaraplateYamlOptions(
                git_remote_type=None, cookiecutter_context_type=SetupCfg
            ),
        ),
    ],
)
def test_get_scaraplate_yaml_options_valid(
    tempdir_path: Path, yaml_text, expected
) -> None:
    (tempdir_path / "scaraplate.yaml").write_text(yaml_text)
    scaraplate_yaml_options = get_scaraplate_yaml_options(tempdir_path)
    assert scaraplate_yaml_options == expected


@pytest.mark.parametrize(
    "yaml_text, cookiecutter_context_dict, expected",
    [
        # Basic syntax:
        (
            """
default_strategy: scaraplate.strategies.Overwrite
strategies_mapping:
  Jenkinsfile: scaraplate.strategies.TemplateHash
  'some/nested/setup.py': scaraplate.strategies.TemplateHash
""",
            {},
            ScaraplateYamlStrategies(
                default_strategy=StrategyNode(
                    strategy=scaraplate.strategies.Overwrite, config={}
                ),
                strategies_mapping={
                    "Jenkinsfile": StrategyNode(
                        strategy=scaraplate.strategies.TemplateHash, config={}
                    ),
                    "some/nested/setup.py": StrategyNode(
                        strategy=scaraplate.strategies.TemplateHash, config={}
                    ),
                },
            ),
        ),
        # Extended syntax (with config):
        (
            """
default_strategy:
  strategy: scaraplate.strategies.Overwrite
  config:
    some_key: True
strategies_mapping:
  other_file.txt:
    strategy: scaraplate.strategies.IfMissing
    unrelated_key_but_okay: 1
  another_file.txt:
    strategy: scaraplate.strategies.SortedUniqueLines
    config:
      some_key: True
""",
            {},
            ScaraplateYamlStrategies(
                default_strategy=StrategyNode(
                    strategy=scaraplate.strategies.Overwrite, config={"some_key": True}
                ),
                strategies_mapping={
                    "other_file.txt": StrategyNode(
                        strategy=scaraplate.strategies.IfMissing, config={}
                    ),
                    "another_file.txt": StrategyNode(
                        strategy=scaraplate.strategies.SortedUniqueLines,
                        config={"some_key": True},
                    ),
                },
            ),
        ),
        # Jinja2 filenames and globs:
        (
            """
default_strategy: scaraplate.strategies.Overwrite
strategies_mapping:
  '*.txt': scaraplate.strategies.IfMissing
  'src/{{ cookiecutter.file1 }}': scaraplate.strategies.TemplateHash
  'src/{{ cookiecutter.file2 }}': scaraplate.strategies.IfMissing
""",
            {"file1": "zzz.py", "file2": "aaa.py"},
            ScaraplateYamlStrategies(
                default_strategy=StrategyNode(
                    strategy=scaraplate.strategies.Overwrite, config={}
                ),
                strategies_mapping={
                    "*.txt": StrategyNode(
                        strategy=scaraplate.strategies.IfMissing, config={}
                    ),
                    "src/aaa.py": StrategyNode(
                        strategy=scaraplate.strategies.IfMissing, config={}
                    ),
                    "src/zzz.py": StrategyNode(
                        strategy=scaraplate.strategies.TemplateHash, config={}
                    ),
                },
            ),
        ),
    ],
)
def test_get_scaraplate_yaml_strategies_valid(
    tempdir_path: Path, yaml_text, cookiecutter_context_dict, expected
) -> None:
    (tempdir_path / "scaraplate.yaml").write_text(yaml_text)
    scaraplate_yaml_strategies = get_scaraplate_yaml_strategies(
        tempdir_path, cookiecutter_context_dict
    )
    assert scaraplate_yaml_strategies == expected


@pytest.mark.parametrize(
    "cls",
    [
        "tempfile.TemporaryDirectory",
        "tempfile",
        "scaraplate.strategies.Strategy",
        '{"strategy": "tempfile.TemporaryDirectory"}',
        '{"strategy": 42}',
        '{"config": {}}',  # strategy is missing
        '{"strategy": "scaraplate.strategies.Overwrite", "config": 42}',
        "42",
    ],
)
@pytest.mark.parametrize("mutation_target", ["default_strategy", "strategies_mapping"])
def test_get_scaraplate_yaml_strategies_invalid(
    tempdir_path: Path, cls: str, mutation_target: str
) -> None:
    classes = dict(
        default_strategy="scaraplate.strategies.Overwrite",
        strategies_mapping="scaraplate.strategies.Overwrite",
    )
    classes[mutation_target] = cls

    yaml_text = f"""
default_strategy: {classes['default_strategy']}
strategies_mapping:
  Jenkinsfile: {classes['strategies_mapping']}
"""
    (tempdir_path / "scaraplate.yaml").write_text(yaml_text)
    with pytest.raises(ValueError):
        get_scaraplate_yaml_strategies(tempdir_path, CookieCutterContextDict({}))


@pytest.mark.parametrize(
    "key, cookiecutter_context_dict",
    [
        ("'{{ cookiecutter.missingkey }}'", {"anotherkey": "42"}),
        ("'{{ somevar }}'", {"somavar": "42"}),  # doesn't start with `cookiecutter.`
    ],
)
def test_get_scaraplate_yaml_strategies_invalid_keys(
    tempdir_path: Path, key, cookiecutter_context_dict
) -> None:
    yaml_text = f"""
default_strategy: scaraplate.strategies.Overwrite
strategies_mapping:
  {key}: scaraplate.strategies.Overwrite
"""
    (tempdir_path / "scaraplate.yaml").write_text(yaml_text)
    with pytest.raises(jinja2.exceptions.UndefinedError):
        get_scaraplate_yaml_strategies(tempdir_path, cookiecutter_context_dict)


@pytest.mark.parametrize(
    "cls",
    [
        "tempfile.TemporaryDirectory",
        "tempfile",
        "scaraplate.gitremotes.GitRemote",
        "scaraplate.cookiecutter.ScaraplateConf",
        '{"strategy": "scaraplate.gitremotes.GitLab"}',
        "42",
    ],
)
def test_get_scaraplate_yaml_options_invalid_git_remotes(
    tempdir_path: Path, cls: str
) -> None:
    yaml_text = f"""
git_remote_type: {cls}
"""
    (tempdir_path / "scaraplate.yaml").write_text(yaml_text)
    with pytest.raises(ValueError):
        get_scaraplate_yaml_options(tempdir_path)


@pytest.mark.parametrize(
    "cls",
    [
        "tempfile.TemporaryDirectory",
        "tempfile",
        "scaraplate.gitremotes.GitHub",
        "scaraplate.cookiecutter.CookieCutterContext",
        '{"strategy": "scaraplate.cookiecutter.ScaraplateConf"}',
        "42",
    ],
)
def test_get_scaraplate_yaml_options_invalid_cookiecutter_context(
    tempdir_path: Path, cls: str
) -> None:
    yaml_text = f"""
cookiecutter_context_type: {cls}
"""
    (tempdir_path / "scaraplate.yaml").write_text(yaml_text)
    with pytest.raises(ValueError):
        get_scaraplate_yaml_options(tempdir_path)



import re

from marshmallow import ValidationError, fields


class Pattern(fields.Field):
    def _serialize(self, value, attr, obj):
        self.fail("not implemented")

    def _deserialize(self, value, attr, data, **kwargs):
        if value is None:
            return None

        try:
            return re.compile(value)
        except re.error as exc:
            raise ValidationError(f"Unable to compile PCRE pattern: {exc}")



from .automation import (
    GitCloneProjectVCS,
    GitCloneTemplateVCS,
    ProjectVCS,
    TemplateVCS,
    automatic_rollup,
)
from .rollup import InvalidScaraplateTemplateError, rollup
from .template import TemplateMeta

__all__ = (
    "GitCloneProjectVCS",
    "GitCloneTemplateVCS",
    "InvalidScaraplateTemplateError",
    "ProjectVCS",
    "TemplateMeta",
    "TemplateVCS",
    "automatic_rollup",
    "rollup",
)



import contextlib
import datetime
import logging
import os
import tempfile
import uuid
from pathlib import Path
from typing import Dict, Iterator, List, Optional
from urllib.parse import urlparse, urlunparse

from pkg_resources import get_distribution

from scaraplate.automation.base import ProjectVCS, TemplateVCS
from scaraplate.config import get_scaraplate_yaml_options
from scaraplate.template import TemplateMeta, _call_git, get_template_meta_from_git

__all__ = ("GitCloneProjectVCS", "GitCloneTemplateVCS")

logger = logging.getLogger("scaraplate")


def scaraplate_version() -> str:
    return get_distribution("scaraplate").version


class GitCloneTemplateVCS(TemplateVCS):
    """A ready to use :class:`.TemplateVCS` implementation which:

    - Uses git
    - Clones a git repo with the template to a temporary directory
      (which is cleaned up afterwards)
    - Allows to specify an inner dir inside the git repo as the template
      root (which is useful for monorepos)
    """

    def __init__(self, template_path: Path, template_meta: TemplateMeta) -> None:
        self._template_path = template_path
        self._template_meta = template_meta

    @property
    def dest_path(self) -> Path:
        return self._template_path

    @property
    def template_meta(self) -> TemplateMeta:
        return self._template_meta

    @classmethod
    @contextlib.contextmanager
    def clone(
        cls,
        clone_url: str,
        *,
        clone_ref: Optional[str] = None,
        monorepo_inner_path: Optional[Path] = None,
    ) -> Iterator["GitCloneTemplateVCS"]:
        """Provides an instance of this class by issuing ``git clone``
        to a tempdir when entering the context manager. Returns a context
        manager object which after ``__enter__`` returns an instance
        of this class.

        :param clone_url: Any valid ``git clone`` url.
        :param clone_ref: Git ref to checkout after clone
            (i.e. branch or tag name).
        :param monorepo_inner_path: Path to the root dir of template
            relative to the root of the repo. If ``None``, the root of
            the repo will be used as the root of template.
        """

        with tempfile.TemporaryDirectory() as tmpdir_name:
            tmpdir_path = Path(tmpdir_name).resolve()
            template_path = tmpdir_path / "scaraplate_template"
            template_path.mkdir()

            git = Git.clone(
                clone_url,
                target_path=template_path,
                ref=clone_ref,
                # We need to strip credentials from the clone_url,
                # because otherwise urls generated for TemplateMeta
                # would contain them, and we don't want that.
                strip_credentials_from_remote=True,
            )
            template_path = git.cwd

            if monorepo_inner_path is not None:
                template_path = template_path / monorepo_inner_path

            scaraplate_yaml_options = get_scaraplate_yaml_options(template_path)
            template_meta = get_template_meta_from_git(
                template_path, git_remote_type=scaraplate_yaml_options.git_remote_type
            )
            if clone_ref is not None:
                assert clone_ref == template_meta.head_ref

            yield cls(template_path, template_meta)


class GitCloneProjectVCS(ProjectVCS):
    """A ready to use :class:`.ProjectVCS` implementation which:

    - Uses git
    - Clones a git repo with the project to a temporary directory
      (which is cleaned up afterwards)
    - Allows to specify an inner dir inside the git repo as the project
      root (which is useful for monorepos)
    - Implements :meth:`.ProjectVCS.commit_changes` as
      ``git commit`` + ``git push``.
    """

    def __init__(
        self,
        project_path: Path,
        git: "Git",
        *,
        changes_branch: str,
        commit_author: str,
        commit_message_template: str,
    ) -> None:
        self._project_path = project_path
        self._git = git
        self.changes_branch = changes_branch
        self.commit_author = commit_author
        self.commit_message_template = commit_message_template
        self.update_time = datetime.datetime.now()

    @property
    def dest_path(self) -> Path:
        return self._project_path

    def is_dirty(self) -> bool:
        return self._git.is_dirty()

    def commit_changes(self, template_meta: TemplateMeta) -> None:
        assert self.is_dirty()

        remote_branch = self._git.remote_ref(self.changes_branch)

        # Create a definitely not existing local branch:
        local_branch = f"{self.changes_branch}{uuid.uuid4()}"
        self._git.checkout_branch(local_branch)

        self._git.commit_all(
            self.format_commit_message(template_meta=template_meta),
            author=self.commit_author,
        )

        if not self._git.is_existing_ref(remote_branch):
            self._git.push(self.changes_branch)
        else:
            # A branch with updates already exists in the remote.

            if self._git.is_same_commit(remote_branch, f"{local_branch}^1"):
                # The `changes_branch` is the same as the clone branch,
                # so essentially the created commit forms a linear history.
                # No need for any diffs here, we just need to push that.
                self._git.push(self.changes_branch)
            else:
                # The two branches have diverged, we need to compare them:
                changes: bool = not self._git.are_one_commit_diffs_equal(
                    local_branch, remote_branch
                )

                if changes:
                    # We could've used force push here, but instead we delete
                    # the branch first, because in GitLab it would also close
                    # the existing MR (if any), and we want that instead of
                    # silently updating the old MR.
                    self._git.push_branch_delete(self.changes_branch)
                    self._git.push(self.changes_branch)
                else:
                    logger.info(
                        "scaraplate did update the project, but there's "
                        "an already existing branch in remote which diff "
                        "is equal to the just produced changes"
                    )

        # Now we should ensure that a Pull Request exists for
        # the `self.changes_branch`, but this class is designed to be agnostic
        # from concrete git remotes, so it should be done in a child class.

    def format_commit_message(self, *, template_meta: TemplateMeta) -> str:
        return self.commit_message_template.format(
            # TODO retrieve path from self.clone_url and pass it here too?
            # (but careful: that clone_url might contain credentials).
            update_time=self.update_time,
            scaraplate_version=scaraplate_version(),
            template_meta=template_meta,
        )

    @classmethod
    @contextlib.contextmanager
    def clone(
        cls,
        clone_url: str,
        *,
        clone_ref: Optional[str] = None,
        monorepo_inner_path: Optional[Path] = None,
        changes_branch: str,
        commit_author: str,
        commit_message_template: str = (
            "Scheduled template update ({update_time:%Y-%m-%d})\n"
            "\n"
            "* scaraplate version: {scaraplate_version}\n"
            "* template commit: {template_meta.commit_url}\n"
            "* template ref: {template_meta.head_ref}\n"
        ),
    ) -> Iterator["GitCloneProjectVCS"]:
        """Provides an instance of this class by issuing ``git clone``
        to a tempdir when entering the context manager. Returns a context
        manager object which after ``__enter__`` returns an instance
        of this class.

        :param clone_url: Any valid ``git clone`` url.
        :param clone_ref: Git ref to checkout after clone
            (i.e. branch or tag name).
        :param monorepo_inner_path: Path to the root dir of project
            relative to the root of the repo. If ``None``, the root of
            the repo will be used as the root of project.
        :param changes_branch: The branch name where the changes should be
            pushed in the remote. Might be the same as ``clone_ref``.
            Note that this branch is never force-pushed. If upon push
            the branch already exists in remote and its one-commit diff
            is different from the one-commit diff of the just created
            local branch, then the remote branch will be deleted and
            the local branch will be pushed to replace the previous one.
        :param commit_author: Author name to use for ``git commit``, e.g.
            ``John Doe <john@example.org>``.
        :param commit_message_template: :meth:`str.format` template
            which is used to produce a commit message when committing
            the changes. Available format variables are:

            - ``update_time`` [:class:`datetime.datetime`] -- the time
              of update
            - ``scaraplate_version`` [:class:`str`] -- scaraplate package
              version
            - ``template_meta`` [:class:`.TemplateMeta`] -- template meta
              returned by :meth:`.TemplateVCS.template_meta`
        """

        with tempfile.TemporaryDirectory() as tmpdir_name:
            tmpdir_path = Path(tmpdir_name).resolve()
            project_path = tmpdir_path / "scaraplate_project"
            project_path.mkdir()

            git = Git.clone(clone_url, target_path=project_path, ref=clone_ref)
            project_path = git.cwd

            if monorepo_inner_path is not None:
                project_path = project_path / monorepo_inner_path

            yield cls(
                project_path,
                git,
                changes_branch=changes_branch,
                commit_author=commit_author,
                commit_message_template=commit_message_template,
            )


class Git:
    def __init__(self, cwd: Path, remote: str = "origin") -> None:
        self.cwd = cwd
        self.remote = remote

    def remote_ref(self, ref: str) -> str:
        return f"{self.remote}/{ref}"

    def checkout_branch(self, branch: str) -> None:
        self._git(["checkout", "-b", branch])

    def commit_all(self, commit_message: str, author: Optional[str] = None) -> None:
        self._git(["add", "--all"])
        extra: List[str] = []
        if author is not None:
            extra = ["--author", author]
        self._git(
            ["commit", "-m", commit_message, *extra],
            env={
                # git would fail if there's no `user.email` in the local
                # git config, even if `--author` is specified.
                "EMAIL": "scaraplate@localhost",
                "GIT_AUTHOR_EMAIL": "scaraplate@localhost",
                "GIT_AUTHOR_NAME": "scaraplate",
                "GIT_COMMITTER_EMAIL": "scaraplate@localhost",
                "GIT_COMMITTER_NAME": "scaraplate",
                "USERNAME": "scaraplate",
            },
        )

    def is_dirty(self) -> bool:
        return bool(self._git(["status", "--porcelain"]))

    def is_existing_ref(self, ref: str) -> bool:
        try:
            self._git(["rev-parse", "--verify", ref])
        except RuntimeError:
            return False
        else:
            return True

    def is_same_commit(self, ref1: str, ref2: str) -> bool:
        commit1 = self._git(["rev-parse", "--verify", ref1])
        commit2 = self._git(["rev-parse", "--verify", ref2])
        return commit1 == commit2

    def are_one_commit_diffs_equal(self, ref1: str, ref2: str) -> bool:
        diff1 = self._git(["diff", f"{ref1}^1..{ref1}"])
        diff2 = self._git(["diff", f"{ref2}^1..{ref2}"])
        return diff1 == diff2

    def push_branch_delete(self, branch: str) -> None:
        self._git(["push", "--delete", self.remote, branch])

    def push(self, ref: str) -> None:
        # https://stackoverflow.com/a/4183856
        self._git(["push", self.remote, f"HEAD:{ref}"])

    def _git(self, args: List[str], *, env: Optional[Dict[str, str]] = None) -> str:
        return _call_git(args, cwd=self.cwd, env=env)

    @classmethod
    def clone(
        cls,
        clone_url: str,
        *,
        target_path: Path,
        ref: Optional[str] = None,
        strip_credentials_from_remote: bool = False,
    ) -> "Git":
        remote = "origin"
        clone_url_without_creds = strip_credentials_from_git_remote(clone_url)

        args = ["clone", clone_url]

        if ref is not None:
            # git-clone(1) explicitly mentions that both branches and tags
            # are allowed in the `--branch`.
            args.extend(["--branch", ref])

        _call_git(args, cwd=target_path)

        actual_items_in_target_path = os.listdir(target_path)
        if len(actual_items_in_target_path) != 1:
            raise RuntimeError(
                f"Expected `git clone` to create exactly one directory. "
                f"Directories in the target: {actual_items_in_target_path}"
            )

        (cloned_dir,) = actual_items_in_target_path
        target_path = target_path / cloned_dir

        if strip_credentials_from_remote:
            _call_git(
                ["remote", "set-url", remote, clone_url_without_creds], cwd=target_path
            )

        return cls(cwd=target_path, remote=remote)


def strip_credentials_from_git_remote(remote: str) -> str:
    parsed = urlparse(remote)
    if not parsed.scheme:
        # Not a URL (probably an SSH remote)
        return remote
    assert parsed.hostname is not None
    clean = parsed._replace(netloc=parsed.hostname)
    return urlunparse(clean)



from .base import ProjectVCS, TemplateVCS, automatic_rollup
from .git import GitCloneProjectVCS, GitCloneTemplateVCS

__all__ = (
    "GitCloneProjectVCS",
    "GitCloneTemplateVCS",
    "ProjectVCS",
    "TemplateVCS",
    "automatic_rollup",
)



import http.server
import socketserver
import threading
from concurrent.futures import Future
from pathlib import Path
from typing import Any, Callable, List, Optional

import pytest


def convert_git_repo_to_bare(call_git, *, cwd: Path) -> None:
    """Git bare repo is a git repo without a working copy. `git clone`
    can clone these repos my simply pointing at their location in the local
    filesystem.
    """
    # https://stackoverflow.com/a/3251126
    call_git("git config --bool core.bare true", cwd=cwd)


@pytest.fixture
def template_bare_git_repo(request, tempdir_path, init_git_and_commit, call_git):
    template_path = tempdir_path / "remote_template"

    cookiecutter_path = template_path / "{{cookiecutter.project_dest}}"
    cookiecutter_path.mkdir(parents=True)
    if request.node.get_closest_marker("template_with_sense_vars") is not None:
        (cookiecutter_path / "sense_vars").write_text("{{ cookiecutter|jsonify }}\n")
    (cookiecutter_path / ".scaraplate.conf").write_text(
        """[cookiecutter_context]
{%- for key, value in cookiecutter.items()|sort %}
{%- if key not in ('_output_dir',) %}
{{ key }} = {{ value }}
{%- endif %}
{%- endfor %}
"""
    )
    (template_path / "cookiecutter.json").write_text(
        '{"project_dest": "test", "key1": null, "key2": null}'
    )
    (template_path / "scaraplate.yaml").write_text(
        """
git_remote_type: scaraplate.gitremotes.GitLab
default_strategy: scaraplate.strategies.Overwrite
strategies_mapping: {}
        """
    )
    init_git_and_commit(template_path, with_remote=False)
    call_git("git branch master2", cwd=template_path)

    convert_git_repo_to_bare(call_git, cwd=template_path)

    return template_path


@pytest.fixture
def project_bare_git_repo(tempdir_path, init_git_and_commit, call_git):
    target_project_path = tempdir_path / "remote_project"

    target_project_path.mkdir(parents=True)
    (target_project_path / "readme").write_text("hi")
    init_git_and_commit(target_project_path, with_remote=False)
    call_git("git branch master2", cwd=target_project_path)

    convert_git_repo_to_bare(call_git, cwd=target_project_path)

    return target_project_path


@pytest.fixture
def http_server():
    server_thread = HttpServerThread()
    server_thread.start()
    yield server_thread
    server_thread.stop()
    server_thread.join()


class HttpServerThread(threading.Thread):
    spinup_timeout = 10

    def __init__(self):
        self.server_host = "127.0.0.1"
        self.server_port = None  # randomly selected by OS

        self.http_server = None
        self._errors: List[Exception] = []
        self.request_handler: Optional[Callable[..., Any]] = None

        # Future is generic. Pylint is wrong.
        # https://github.com/python/typeshed/blob/6e4708ebf3a482aa8d524196712c37c9fd645953/stdlib/3/concurrent/futures/_base.pyi#L26  # noqa
        self.socket_created_future: Future[  # pylint: disable=unsubscriptable-object
            bool
        ] = Future()

        super().__init__()
        self.daemon = True

    def set_request_handler(
        self, request_handler: Callable[..., Any]
    ) -> Callable[..., Any]:
        assert self.request_handler is None
        self.request_handler = request_handler
        return request_handler

    def get_url(self):
        assert self.socket_created_future.result(self.spinup_timeout)
        return f"http://{self.server_host}:{self.server_port}"

    def run(self):
        assert (
            self.http_server is None
        ), "This class is not reentrable. Please create a new instance."

        thread = self

        class Server(http.server.BaseHTTPRequestHandler):
            def __request_handler(self, method):
                if thread.request_handler is None:  # pragma: no cover
                    self.send_error(500, "request_handler is None")
                else:
                    try:
                        # https://stackoverflow.com/a/20879937
                        body = self.rfile.read(
                            int(self.headers.get("content-length", 0))
                        )
                        (code, headers, body,) = thread.request_handler(
                            method, self.path, headers=self.headers, body=body
                        )
                        self.send_response(code)
                        for name, value in headers.items():
                            self.send_header(name, value)
                        self.end_headers()
                        self.wfile.write(body)
                    except Exception as e:  # pragma: no cover
                        thread._errors.append(e)

            def do_HEAD(self):  # pragma: no cover
                self.__request_handler("HEAD")

            def do_GET(self):
                self.__request_handler("GET")

            def do_POST(self):  # pragma: no cover
                self.__request_handler("POST")

            def do_PATCH(self):  # pragma: no cover
                self.__request_handler("PATCH")

        # ThreadingTCPServer offloads connections to separate threads, so
        # the serve_forever loop doesn't block until connection is closed
        # (unlike TCPServer). This allows to shutdown the serve_forever loop
        # even if there's an open connection.
        try:
            # TODO: switch to http.server.ThreadingHTTPServer
            # when the py36 support would be dropped
            self.http_server = socketserver.ThreadingTCPServer(
                (self.server_host, 0), Server
            )

            # don't hang if there're some open connections
            self.http_server.daemon_threads = True

            self.server_port = self.http_server.server_address[1]
        except Exception as e:  # pragma: no cover
            self.socket_created_future.set_exception(e)
            raise
        else:
            self.socket_created_future.set_result(True)

        self.http_server.serve_forever()

    def stop(self):
        assert self.http_server is not None
        self.http_server.shutdown()  # stop serve_forever()
        self.http_server.server_close()

        # assert not self._errors
        # ^^^ gives unreadable assertions in pytest
        # This is better:
        if self._errors:
            raise self._errors[0]



from pkg_resources import get_distribution


def test_metadata():
    metadata_name = "scaraplate"
    pkg = get_distribution(metadata_name)
    assert pkg.version
    assert pkg.project_name == metadata_name


def test_package_import():
    import scaraplate

    assert scaraplate



"""Scaraplate assumes that the template dir is a git repo.

Strategies receive a :class:`scaraplate.template.TemplateMeta` instance
which contains URLs to the template's project and the HEAD git commit
on a git remote's web interface (such as GitHub). These URLs might be
rendered in the target files by the strategies.

Scaraplate has built-in support for some popular git remotes. The remote
is attempted to be detected automatically, but if that fails, it should
be specified manually.


Sample ``scaraplate.yaml`` excerpt:

::

    git_remote_type: scaraplate.gitremotes.GitHub

"""
import abc
import re
from typing import Optional, Type


def _dot_git_remote_to_https(remote_url: str) -> str:
    url = remote_url
    # `git@gitlab.com:<suffix>` -> `https://gitlab.com/<suffix>`
    url = re.sub(r"^[^@]*@([^:]+):", r"https://\1/", url)
    url = re.sub(r".git$", "", url)
    return url


def make_git_remote(
    remote: str, *, git_remote_type: Optional[Type["GitRemote"]] = None
) -> "GitRemote":
    if git_remote_type is not None:
        return git_remote_type(remote)

    if "gitlab" in remote.lower():
        return GitLab(remote)
    elif "github" in remote.lower():
        return GitHub(remote)
    elif "bitbucket" in remote.lower():
        return BitBucket(remote)
    else:
        raise ValueError(
            "Unable to automatically determine the GitRemote type. "
            "Please set a specific one in the `scaraplate.yaml` config "
            "using the `git_remote_type` option."
        )


class GitRemote(abc.ABC):
    """Base class for a git remote implementation, which generates http
    URLs from a git remote (either ssh of http) and a commit hash.
    """

    def __init__(self, remote: str) -> None:
        """Init the git remote.

        :param remote: A git remote, either ssh or http(s).
        """
        self.remote = remote

    @abc.abstractmethod
    def project_url(self) -> str:
        """Return a project URL at the given git remote."""
        pass

    @abc.abstractmethod
    def commit_url(self, commit_hash: str) -> str:
        """Return a commit URL at the given git remote.

        :param commit_hash: Git commit hash.
        """
        pass


class GitLab(GitRemote):
    """GitLab git remote implementation."""

    def project_url(self) -> str:
        return _dot_git_remote_to_https(self.remote)

    def commit_url(self, commit_hash: str) -> str:
        project_url = self.project_url()
        return f"{project_url.rstrip('/')}/commit/{commit_hash}"


class GitHub(GitRemote):
    """GitHub git remote implementation."""

    def project_url(self) -> str:
        return _dot_git_remote_to_https(self.remote)

    def commit_url(self, commit_hash: str) -> str:
        project_url = self.project_url()
        return f"{project_url.rstrip('/')}/commit/{commit_hash}"


class BitBucket(GitRemote):
    """BitBucket git remote implementation."""

    def project_url(self) -> str:
        return _dot_git_remote_to_https(self.remote)

    def commit_url(self, commit_hash: str) -> str:
        project_url = self.project_url()
        return f"{project_url.rstrip('/')}/commits/{commit_hash}"



import contextlib
import json
import os
from pathlib import Path
from typing import Dict, Optional
from unittest.mock import ANY, sentinel

import pytest

from scaraplate.config import ScaraplateYamlOptions, ScaraplateYamlStrategies
from scaraplate.cookiecutter import ScaraplateConf
from scaraplate.rollup import (
    InvalidScaraplateTemplateError,
    get_project_dest,
    get_strategy,
    get_target_project_cookiecutter_context,
    get_template_root_and_dir,
    rollup,
)


@contextlib.contextmanager
def with_working_directory(target_dir: Path):
    cur = os.getcwd()
    os.chdir(target_dir)
    try:
        yield
    finally:
        os.chdir(cur)


@pytest.mark.parametrize("apply_count", [1, 2])
def test_rollup_fuzzy(tempdir_path, apply_count, init_git_and_commit):
    template_path = tempdir_path / "template"
    target_project_path = tempdir_path / "test"

    # Prepare template
    cookiecutter_path = template_path / "{{cookiecutter.project_dest}}"
    cookiecutter_path.mkdir(parents=True)
    (cookiecutter_path / "README.md").write_text(
        "{{ cookiecutter.project_dest }} mock!"
    )
    (cookiecutter_path / "setup.py").write_text("#!/usr/bin/env python\n")
    (cookiecutter_path / "setup.py").chmod(0o755)
    (cookiecutter_path / "sense_vars").write_text("{{ cookiecutter|jsonify }}\n")
    (cookiecutter_path / ".scaraplate.conf").write_text(
        """[cookiecutter_context]
{%- for key, value in cookiecutter.items()|sort %}
{%- if key not in ('_output_dir',) %}
{{ key }} = {{ value }}
{%- endif %}
{%- endfor %}
"""
    )
    (template_path / "cookiecutter.json").write_text('{"project_dest": "test"}')
    (template_path / "scaraplate.yaml").write_text(
        """
default_strategy: scaraplate.strategies.Overwrite
strategies_mapping:
  setup.py: scaraplate.strategies.TemplateHash
        """
    )
    init_git_and_commit(template_path)

    # Apply template (possibly multiple times)
    for i in range(apply_count):
        rollup(
            template_dir=str(template_path),
            target_project_dir=str(target_project_path),
            no_input=True,
        )

        assert "test mock!" == (target_project_path / "README.md").read_text()
        assert 0o755 == (0o777 & (target_project_path / "setup.py").stat().st_mode)

        with open((target_project_path / "sense_vars"), "rt") as f:
            assert json.load(f) == {
                "_output_dir": ANY,
                "_template": "template",
                "project_dest": "test",
            }


def test_add_remove_template_var(tempdir_path, init_git_and_commit):
    template_path = tempdir_path / "template"
    target_project_path = tempdir_path / "test"

    # Prepare template
    cookiecutter_path = template_path / "{{cookiecutter.project_dest}}"
    cookiecutter_path.mkdir(parents=True)
    (cookiecutter_path / ".scaraplate.conf").write_text(
        """[cookiecutter_context]
{%- for key, value in cookiecutter.items()|sort %}
{%- if key not in ('_output_dir',) %}
{{ key }} = {{ value }}
{%- endif %}
{%- endfor %}
"""
    )
    (template_path / "cookiecutter.json").write_text(
        '{"project_dest": "test", "removed_var": 42}'
    )
    (template_path / "scaraplate.yaml").write_text(
        """
default_strategy: scaraplate.strategies.Overwrite
strategies_mapping: {}
        """
    )
    init_git_and_commit(template_path)

    rollup(
        template_dir=str(template_path),
        target_project_dir=str(target_project_path),
        no_input=True,
    )
    assert (target_project_path / ".scaraplate.conf").read_text() == (
        """[cookiecutter_context]
_template = template
project_dest = test
removed_var = 42
"""
    )

    # Remove `removed_var` and add `added_var`
    (template_path / "cookiecutter.json").write_text(
        '{"project_dest": "test", "added_var": 24}'
    )
    rollup(
        template_dir=str(template_path),
        target_project_dir=str(target_project_path),
        no_input=True,
    )
    assert (target_project_path / ".scaraplate.conf").read_text() == (
        """[cookiecutter_context]
_template = template
added_var = 24
project_dest = test
"""
    )


@pytest.mark.parametrize("create_empty", [False, True])
def test_invalid_template_is_raised_for_missing_cookiecutter_context(
    create_empty, tempdir_path, init_git_and_commit
):
    template_path = tempdir_path / "template"
    target_project_path = tempdir_path / "test"

    # Prepare template
    cookiecutter_path = template_path / "{{cookiecutter.project_dest}}"
    cookiecutter_path.mkdir(parents=True)
    if create_empty:
        (cookiecutter_path / ".scaraplate.conf").write_text("")
    (template_path / "cookiecutter.json").write_text('{"project_dest": "test"}')
    (template_path / "scaraplate.yaml").write_text(
        """
default_strategy: scaraplate.strategies.Overwrite
strategies_mapping: {}
        """
    )
    init_git_and_commit(template_path)

    with pytest.raises(InvalidScaraplateTemplateError) as excinfo:
        rollup(
            template_dir=str(template_path),
            target_project_dir=str(target_project_path),
            no_input=True,
        )
    assert "cookiecutter context file `.scaraplate.conf` " in str(excinfo.value)


def test_extra_context(tempdir_path, init_git_and_commit):
    template_path = tempdir_path / "template"
    target_project_path = tempdir_path / "test"

    # Prepare template
    cookiecutter_path = template_path / "{{cookiecutter.project_dest}}"
    cookiecutter_path.mkdir(parents=True)
    (cookiecutter_path / "sense_vars").write_text("{{ cookiecutter|jsonify }}\n")
    (cookiecutter_path / ".scaraplate.conf").write_text(
        """[cookiecutter_context]
{%- for key, value in cookiecutter.items()|sort %}
{%- if key not in ('_output_dir',) %}
{{ key }} = {{ value }}
{%- endif %}
{%- endfor %}
"""
    )
    (template_path / "cookiecutter.json").write_text(
        '{"project_dest": "test", "key1": null, "key2": null}'
    )
    (template_path / "scaraplate.yaml").write_text(
        """
default_strategy: scaraplate.strategies.Overwrite
strategies_mapping: {}
        """
    )
    init_git_and_commit(template_path)

    # Initial rollup
    rollup(
        template_dir=str(template_path),
        target_project_dir=str(target_project_path),
        no_input=True,
        extra_context={"key1": "initial1", "key2": "initial2"},
    )
    with open((target_project_path / "sense_vars"), "rt") as f:
        assert json.load(f) == {
            "_output_dir": ANY,
            "_template": "template",
            "project_dest": "test",
            "key1": "initial1",
            "key2": "initial2",
        }

    # A second rollup with a different context
    rollup(
        template_dir=str(template_path),
        target_project_dir=str(target_project_path),
        no_input=True,
        extra_context={"key1": "second1", "key2": "second2"},
    )
    with open((target_project_path / "sense_vars"), "rt") as f:
        assert json.load(f) == {
            "_output_dir": ANY,
            "_template": "template",
            "project_dest": "test",
            "key1": "second1",
            "key2": "second2",
        }


def test_rollup_with_jinja2_mapping(tempdir_path, init_git_and_commit):
    template_path = tempdir_path / "template"
    target_project_path = tempdir_path / "test"

    # Prepare template
    cookiecutter_path = template_path / "{{cookiecutter.project_dest}}"
    cookiecutter_path.mkdir(parents=True)
    (cookiecutter_path / "sense_vars").write_text("{{ cookiecutter|jsonify }}\n")
    (cookiecutter_path / ".scaraplate.conf").write_text(
        """[cookiecutter_context]
{%- for key, value in cookiecutter.items()|sort %}
{%- if key not in ('_output_dir',) %}
{{ key }} = {{ value }}
{%- endif %}
{%- endfor %}
"""
    )
    (template_path / "cookiecutter.json").write_text(
        '{"project_dest": "test", "file1": null, "file2": "boop"}'
    )
    (template_path / "scaraplate.yaml").write_text(
        """
default_strategy: scaraplate.strategies.SortedUniqueLines
strategies_mapping:
  '{{ cookiecutter.file1 }}.txt': scaraplate.strategies.IfMissing
  '{{ cookiecutter.file2 }}.txt': scaraplate.strategies.Overwrite
        """
    )
    (cookiecutter_path / "{{ cookiecutter.file1 }}.txt").write_text("template!")
    (cookiecutter_path / "{{ cookiecutter.file2 }}.txt").write_text("template!")
    init_git_and_commit(template_path)

    # Initial rollup
    rollup(
        template_dir=str(template_path),
        target_project_dir=str(target_project_path),
        no_input=True,
        extra_context={"file1": "beep"},
    )
    assert "template!" == (target_project_path / "beep.txt").read_text()
    assert "template!" == (target_project_path / "boop.txt").read_text()
    (target_project_path / "beep.txt").write_text("target!")
    (target_project_path / "boop.txt").write_text("target!")

    # A second rollup (beep.txt should not be changed)
    rollup(
        template_dir=str(template_path),
        target_project_dir=str(target_project_path),
        no_input=True,
        extra_context={"file1": "beep"},
    )
    assert "target!" == (target_project_path / "beep.txt").read_text()
    assert "template!" == (target_project_path / "boop.txt").read_text()


def test_get_project_dest(tempdir_path: Path) -> None:
    target = tempdir_path / "myproject"
    with with_working_directory(tempdir_path):
        assert "myproject" == get_project_dest(Path("myproject"))

    target.mkdir()
    with with_working_directory(target):
        assert "myproject" == get_project_dest(Path("."))


def test_get_template_root_and_dir(tempdir_path: Path) -> None:
    target = tempdir_path / "myproject"
    target.mkdir()

    with with_working_directory(tempdir_path):
        assert (tempdir_path, "myproject") == get_template_root_and_dir(
            Path("myproject")
        )
        assert (tempdir_path, "myproject") == get_template_root_and_dir(
            tempdir_path / "myproject"
        )

    with with_working_directory(target):
        assert (tempdir_path, "myproject") == get_template_root_and_dir(Path("."))


@pytest.mark.parametrize(
    "contents, expected_context",
    [
        (None, {}),
        ("", {}),
        (
            """
[cookiecutter_context]
metadata_author = Usermodel @ Rambler&Co
coverage_fail_under = 90
project_monorepo_name =
""",
            {
                "metadata_author": "Usermodel @ Rambler&Co",
                "coverage_fail_under": "90",
                "project_monorepo_name": "",
            },
        ),
    ],
)
def test_get_target_project_cookiecutter_context(
    tempdir_path: Path, contents: Optional[str], expected_context: Dict
) -> None:
    if contents is not None:
        (tempdir_path / ".scaraplate.conf").write_text(contents)

    scaraplate_yaml_options = ScaraplateYamlOptions(
        git_remote_type=None, cookiecutter_context_type=ScaraplateConf
    )

    assert expected_context == get_target_project_cookiecutter_context(
        tempdir_path, scaraplate_yaml_options
    )


def test_get_strategy():
    scaraplate_yaml_strategies = ScaraplateYamlStrategies(
        default_strategy=sentinel.default,
        strategies_mapping={
            "Jenkinsfile": sentinel.jenkinsfile,
            "some/nested/setup.py": sentinel.nested_setup_py,
            "src/*/__init__.py": sentinel.glob_init,
        },
    )

    assert sentinel.default is get_strategy(scaraplate_yaml_strategies, Path("readme"))
    assert sentinel.jenkinsfile is get_strategy(
        scaraplate_yaml_strategies, Path("Jenkinsfile")
    )
    assert sentinel.nested_setup_py is get_strategy(
        scaraplate_yaml_strategies, Path("some/nested/setup.py")
    )
    assert sentinel.glob_init is get_strategy(
        scaraplate_yaml_strategies, Path("src/my_project/__init__.py")
    )



import os
import subprocess
from pathlib import Path
from typing import Dict, NamedTuple, Optional, Sequence, Type

from scaraplate.gitremotes import GitRemote, make_git_remote


class TemplateMeta(NamedTuple):
    """Metadata of the template's git repo status."""

    git_project_url: str
    commit_hash: str
    commit_url: str
    is_git_dirty: bool
    head_ref: Optional[str]


def get_template_meta_from_git(
    template_path: Path, *, git_remote_type: Optional[Type[GitRemote]] = None
) -> TemplateMeta:
    remote_url = _git_remote_origin(template_path)
    commit_hash = _git_head_commit_hash(template_path)

    git_remote = make_git_remote(remote_url, git_remote_type=git_remote_type)

    return TemplateMeta(
        git_project_url=git_remote.project_url(),
        commit_hash=commit_hash,
        commit_url=git_remote.commit_url(commit_hash),
        is_git_dirty=_is_git_dirty(template_path),
        head_ref=_git_resolve_head(template_path),
    )


def _git_head_commit_hash(cwd: Path) -> str:
    return _call_git(["rev-parse", "--verify", "HEAD"], cwd)


def _is_git_dirty(cwd: Path) -> bool:
    return bool(_call_git(["status", "--porcelain"], cwd))


def _git_resolve_head(cwd: Path) -> Optional[str]:
    ref = _call_git(["rev-parse", "--abbrev-ref", "HEAD"], cwd)
    if ref == "HEAD":
        # Detached HEAD (some specific commit has been checked out).
        return None
    return ref


def _git_remote_origin(cwd: Path) -> str:
    # Would raise if there's no remote called `origin`.
    return _call_git(["config", "--get", "remote.origin.url"], cwd)


def _call_git(
    command: Sequence[str], cwd: Path, *, env: Optional[Dict[str, str]] = None
) -> str:
    env_combined = dict(os.environ)
    if env:
        env_combined.update(env)

    try:
        out = subprocess.run(
            ["git", *command],
            cwd=cwd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
            env=env_combined,
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"{command} command failed in the directory "
            f"'{cwd}'. "
            f"Ensure that it is a valid git repo."
            f"\n"
            f"stdout:\n"
            f"{e.stdout.decode()}\n"
            f"stderr:\n"
            f"{e.stderr.decode()}\n"
        )
    else:
        return out.stdout.decode().strip()



try:
    from marshmallow import __version_info__

    is_marshmallow_3 = __version_info__[0] >= 3
except ImportError:  # pragma: no cover
    is_marshmallow_3 = False


def marshmallow_load_data(schema, data):
    if is_marshmallow_3:
        return schema().load(data)
    else:  # pragma: no cover
        # 2.X line
        return schema(strict=True).load(data).data


def marshmallow_pass_original_for_many(original_data, many):
    if is_marshmallow_3:
        return [original_data]
    else:  # pragma: no cover
        if not many:
            # `many=True` field would contain a list here, otherwise
            # it would be a dict.
            original_data = [original_data]
        return original_data



import abc
import logging
from contextlib import ExitStack
from pathlib import Path
from typing import ContextManager, Mapping, Optional

from scaraplate.rollup import rollup
from scaraplate.template import TemplateMeta

logger = logging.getLogger("scaraplate")

__all__ = ("TemplateVCS", "ProjectVCS", "automatic_rollup")


def automatic_rollup(
    *,
    template_vcs_ctx: ContextManager["TemplateVCS"],
    project_vcs_ctx: ContextManager["ProjectVCS"],
    extra_context: Optional[Mapping[str, str]] = None
) -> None:
    """The main function of the automated rollup implementation.

    This function accepts two context managers, which should return
    two classes: :class:`.TemplateVCS` and :class:`.ProjectVCS`,
    which represent the cloned template and target project
    correspondingly.

    The context managers should prepare the repos, e.g. they should
    create a temporary directory, clone a repo there, and produce
    a :class:`.TemplateVCS` or :class:`.ProjectVCS` class instance.

    This function then applies scaraplate rollup of the template
    to the target project in :ref:`no-input mode <no_input_mode>`.
    If the target project contains any changes (as reported by
    :meth:`.ProjectVCS.is_dirty`), they will be committed by calling
    :meth:`.ProjectVCS.commit_changes`.

    .. versionadded:: 0.2
    """

    with ExitStack() as stack:
        # Clone target project and template
        project_vcs = stack.enter_context(project_vcs_ctx)
        template_vcs = stack.enter_context(template_vcs_ctx)

        rollup(
            template_dir=template_vcs.dest_path,
            target_project_dir=project_vcs.dest_path,
            no_input=True,
            extra_context=extra_context,
        )

        if not project_vcs.is_dirty():
            logger.info(
                "scaraplate rollup didn't change anything -- the project "
                "is in sync with template"
            )
            return

        logger.info("scaraplate rollup produced some changes -- committing them...")
        project_vcs.commit_changes(template_vcs.template_meta)
        logger.info("scaraplate changes have been committed successfully")


class TemplateVCS(abc.ABC):
    """A base class representing a template retrieved from a VCS
    (probably residing in a temporary directory).

    The resulting directory with template must be within a git repository,
    see :doc:`template` for details. But it doesn't mean that it must
    be retrieved from git. Template might be retrieved from anywhere,
    it just has to be in git at the end. That git repo will be used
    to fill the :class:`.TemplateMeta` structure.
    """

    @property
    @abc.abstractmethod
    def dest_path(self) -> Path:
        """Path to the root directory of the template."""
        pass

    @property
    @abc.abstractmethod
    def template_meta(self) -> TemplateMeta:
        """:class:`.TemplateMeta` filled using the template's git repo."""
        pass


class ProjectVCS(abc.ABC):
    """A base class representing a project retrieved from a VCS
    (probably residing in a temporary directory).

    The project might use any VCS, at this point there're no assumptions
    made by scaraplate about the VCS.
    """

    @property
    @abc.abstractmethod
    def dest_path(self) -> Path:
        """Path to the root directory of the project."""
        pass

    @abc.abstractmethod
    def is_dirty(self) -> bool:
        """Tell whether the project has any changes not committed
        to the VCS."""
        pass

    @abc.abstractmethod
    def commit_changes(self, template_meta: TemplateMeta) -> None:
        """Commit the changes made to the project. This method is
        responsible for delivering the changes back to the place
        the project was retrieved from. For example, if the project
        is using ``git`` and it was cloned to a temporary directory,
        then this method should commit the changes and push them back
        to git remote.

        This method will be called only if :meth:`.ProjectVCS.is_dirty`
        has returned True.
        """
        pass



"""cookiecutter context are the variables specified in ``cookiecutter.json``,
which should be provided to cookiecutter to cut a project from the template.

The context should be generated by one of the files in the template,
so scaraplate could read these variables and rollup the template automatically
(i.e. without asking for these variables).

The default context reader is :class:`.ScaraplateConf`, but a custom one
might be specified in ``scaraplate.yaml`` like this:

::

    cookiecutter_context_type: scaraplate.cookiecutter.SetupCfg
"""
import abc
from configparser import ConfigParser
from pathlib import Path
from typing import Dict, NewType

CookieCutterContextDict = NewType("CookieCutterContextDict", Dict[str, str])


def _configparser_from_path(cfg_path: Path) -> ConfigParser:
    parser = ConfigParser()
    text = cfg_path.read_text()
    parser.read_string(text, source=str(cfg_path))
    return parser


class CookieCutterContext(abc.ABC):
    """The abstract base class for retrieving cookiecutter context from
    the target project.

    This class can be extended to provide a custom implementation of
    the context reader.
    """

    def __init__(self, target_path: Path) -> None:
        """Init the context reader."""
        self.target_path = target_path

    @abc.abstractmethod
    def read(self) -> CookieCutterContextDict:
        """Retrieve the context.

        If the target file doesn't exist, :class:`FileNotFoundError`
        must be raised.

        If the file doesn't contain the context, an empty dict
        should be returned.
        """
        pass


class ScaraplateConf(CookieCutterContext):
    """A default context reader which assumes that the cookiecutter
    template contains the following file named ``.scaraplate.conf``
    in the root of the project dir:

    ::

        [cookiecutter_context]
        {%- for key, value in cookiecutter.items()|sort %}
        {%- if key not in ('_output_dir',) %}
        {{ key }} = {{ value }}
        {%- endif %}
        {%- endfor %}

    Cookiecutter context would be rendered in the target project by this
    file, and this class is able to retrieve that context from it.
    """

    section_name = "cookiecutter_context"

    def __init__(self, target_path: Path) -> None:
        super().__init__(target_path)
        self.scaraplate_conf = target_path / ".scaraplate.conf"

    def read(self) -> CookieCutterContextDict:
        configparser = _configparser_from_path(self.scaraplate_conf)
        context_configparser = dict(configparser).get(self.section_name)
        context: Dict[str, str] = dict(context_configparser or {})
        return CookieCutterContextDict(context)

    def __str__(self):
        return f"{self.scaraplate_conf}"


class SetupCfg(CookieCutterContext):
    """A context reader which retrieves the cookiecutter context from
    a section in ``setup.cfg`` file.

    The ``setup.cfg`` file must be in the cookiecutter template and must
    contain the following section:

    ::

        [tool:cookiecutter_context]
        {%- for key, value in cookiecutter.items()|sort %}
        {%- if key not in ('_output_dir',) %}
        {{ key }} = {{ value }}
        {%- endif %}
        {%- endfor %}
    """

    section_name = "tool:cookiecutter_context"

    def __init__(self, target_path: Path) -> None:
        super().__init__(target_path)
        self.setup_cfg = target_path / "setup.cfg"

    def read(self) -> CookieCutterContextDict:
        configparser = _configparser_from_path(self.setup_cfg)
        context_configparser = dict(configparser).get(self.section_name)
        context: Dict[str, str] = dict(context_configparser or {})
        return CookieCutterContextDict(context)

    def __str__(self):
        return f"{self.setup_cfg}"



import pytest

from scaraplate.gitremotes import BitBucket, GitHub, GitLab, make_git_remote


@pytest.mark.parametrize(
    "remote, expected_remote",
    [
        ("git@gitlab.com:pycqa/flake8.git", GitLab),
        ("https://gitlab.com/pycqa/flake8.git", GitLab),
        ("https://gitlab.example.org/pycqa/flake8.git", GitLab),
        ("git@github.com:geopy/geopy.git", GitHub),
        ("https://github.com/geopy/geopy.git", GitHub),
        ("https://github.example.org/geopy/geopy.git", GitHub),
        ("https://bitbucket.org/someuser/someproject.git", BitBucket),
        ("https://git.example.org/pycqa/flake8.git", None),
    ],
)
def test_make_git_remote_detection(remote, expected_remote):
    if expected_remote is None:
        with pytest.raises(ValueError):
            make_git_remote(remote)
    else:
        assert expected_remote is type(make_git_remote(remote))  # noqa


@pytest.mark.parametrize(
    "remote, git_remote_type, expected_remote",
    [("https://git.example.org/pycqa/flake8.git", GitLab, GitLab)],
)
def test_make_git_remote_custom(remote, git_remote_type, expected_remote):
    assert expected_remote is type(  # noqa
        make_git_remote(remote, git_remote_type=git_remote_type)
    )


@pytest.mark.parametrize(
    "remote, expected",
    [
        ("git@gitlab.com:pycqa/flake8.git", "https://gitlab.com/pycqa/flake8"),
        ("https://gitlab.com/pycqa/flake8.git", "https://gitlab.com/pycqa/flake8"),
        (
            "git@gitlab.example.org:pycqa/flake8.git",
            "https://gitlab.example.org/pycqa/flake8",
        ),
        (
            "https://gitlab.example.org/pycqa/flake8.git",
            "https://gitlab.example.org/pycqa/flake8",
        ),
    ],
)
def test_gitlab_url_from_remote(remote, expected):
    assert expected == GitLab(remote).project_url()


@pytest.mark.parametrize(
    "remote, commit_hash, expected_url",
    [
        (
            "https://gitlab.com/pycqa/flake8.git",
            "1111111111111111111111111111111111111111",
            "https://gitlab.com/pycqa/flake8/commit/"
            "1111111111111111111111111111111111111111",
        )
    ],
)
def test_gitlab_commit_url(remote, commit_hash, expected_url):
    assert expected_url == GitLab(remote).commit_url(commit_hash)


@pytest.mark.parametrize(
    "remote, expected",
    [
        ("git@github.com:geopy/geopy.git", "https://github.com/geopy/geopy"),
        ("https://github.com/geopy/geopy.git", "https://github.com/geopy/geopy"),
        (
            "git@github.example.org:geopy/geopy.git",
            "https://github.example.org/geopy/geopy",
        ),
        (
            "https://github.example.org/geopy/geopy.git",
            "https://github.example.org/geopy/geopy",
        ),
    ],
)
def test_github_url_from_remote(remote, expected):
    assert expected == GitHub(remote).project_url()


@pytest.mark.parametrize(
    "remote, commit_hash, expected_url",
    [
        (
            "https://github.com/geopy/geopy.git",
            "1111111111111111111111111111111111111111",
            "https://github.com/geopy/geopy/commit/"
            "1111111111111111111111111111111111111111",
        )
    ],
)
def test_github_commit_url(remote, commit_hash, expected_url):
    assert expected_url == GitHub(remote).commit_url(commit_hash)


@pytest.mark.parametrize(
    "remote, expected",
    [
        (
            "git@bitbucket.org:someuser/someproject.git",
            "https://bitbucket.org/someuser/someproject",
        ),
        (
            "https://bitbucket.org/someuser/someproject.git",
            "https://bitbucket.org/someuser/someproject",
        ),
    ],
)
def test_bitbucket_url_from_remote(remote, expected):
    assert expected == BitBucket(remote).project_url()


@pytest.mark.parametrize(
    "remote, commit_hash, expected_url",
    [
        (
            "https://bitbucket.org/someuser/someproject.git",
            "1111111111111111111111111111111111111111",
            "https://bitbucket.org/someuser/someproject/commits/"
            "1111111111111111111111111111111111111111",
        )
    ],
)
def test_bitbucket_commit_url(remote, commit_hash, expected_url):
    assert expected_url == BitBucket(remote).commit_url(commit_hash)



import contextlib
import logging
from pathlib import Path
from typing import Iterator, Optional
from urllib.parse import urljoin, urlparse, urlunparse

from scaraplate.automation.base import ProjectVCS, TemplateVCS
from scaraplate.automation.git import (
    GitCloneProjectVCS,
    GitCloneTemplateVCS,
    scaraplate_version,
)
from scaraplate.template import TemplateMeta

try:
    import gitlab
except ImportError:
    gitlab_available = False
else:
    gitlab_available = True


logger = logging.getLogger("scaraplate")


def ensure_gitlab_is_installed():
    if not gitlab_available:
        raise ImportError(
            "python-gitlab must be installed in order to use GitLab integration. "
            'Install with `pip install "scaraplate[gitlab]"`.'
        )


def gitlab_clone_url(project_url: str, private_token: Optional[str]) -> str:
    parsed = urlparse(project_url)
    assert parsed.scheme

    if private_token:
        clean = parsed._replace(netloc=f"oauth2:{private_token}@{parsed.hostname}")
    else:
        clean = parsed
    if not clean.path.endswith(".git"):
        clean = clean._replace(path=f"{clean.path}.git")
    return urlunparse(clean)


def gitlab_project_url(gitlab_url: str, full_project_name: str) -> str:
    return urljoin(gitlab_url, full_project_name)


class GitLabCloneTemplateVCS(TemplateVCS):
    """A class which extends :class:`.GitCloneTemplateVCS` with
    GitLab-specific ``clone_url`` generation.
    """

    def __init__(self, git_clone: GitCloneTemplateVCS) -> None:
        self._git_clone = git_clone

    @property
    def dest_path(self) -> Path:
        return self._git_clone.dest_path

    @property
    def template_meta(self) -> TemplateMeta:
        return self._git_clone.template_meta

    @classmethod
    @contextlib.contextmanager
    def clone(
        cls,
        project_url: str,
        private_token: Optional[str] = None,
        *,
        clone_ref: Optional[str] = None,
        monorepo_inner_path: Optional[Path] = None,
    ) -> Iterator["GitLabCloneTemplateVCS"]:
        """Same as :meth:`.GitCloneTemplateVCS.clone` except that
        ``clone_url`` is replaced with ``project_url`` and ``private_token``.

        The ``private_token`` allows to clone private repos, which are
        visible only for an authenticated user.

        :param project_url: A URL to a GitLab project, e.g.
            ``https://gitlab.example.org/myorganization/myproject``.
        :param private_token: GitLab access token,
            see `<https://docs.gitlab.com/ce/api/#oauth2-tokens>`_.
        """

        with GitCloneTemplateVCS.clone(
            clone_url=gitlab_clone_url(project_url, private_token),
            clone_ref=clone_ref,
            monorepo_inner_path=monorepo_inner_path,
        ) as git_clone:
            yield cls(git_clone)


class GitLabMRProjectVCS(ProjectVCS):
    """A class which extends :class:`.GitCloneProjectVCS` with
    GitLab-specific ``clone_url`` generation and opens a GitLab Merge Request
    after ``git push``.
    """

    def __init__(
        self,
        git_clone: GitCloneProjectVCS,
        *,
        gitlab_project,
        mr_title_template: str,
        mr_description_markdown_template: str,
    ) -> None:
        self._git_clone = git_clone
        self._gitlab_project = gitlab_project
        self.mr_title_template = mr_title_template
        self.mr_description_markdown_template = mr_description_markdown_template

    @property
    def dest_path(self) -> Path:
        return self._git_clone.dest_path

    def is_dirty(self) -> bool:
        return self._git_clone.is_dirty()

    def commit_changes(self, template_meta: TemplateMeta) -> None:
        self._git_clone.commit_changes(template_meta)

        if self._git_clone.changes_branch == self._gitlab_project.default_branch:
            logger.info(
                "Skipping MR creation step as `changes_branch` and `default_branch` "
                "are the same, i.e. the changes are already in the target branch."
            )
        else:
            self.create_merge_request(
                title=self.format_merge_request_title(template_meta=template_meta),
                description=self.format_merge_request_description(
                    template_meta=template_meta
                ),
            )

    def format_merge_request_title(self, *, template_meta: TemplateMeta) -> str:
        return self.mr_title_template.format(
            update_time=self._git_clone.update_time, template_meta=template_meta
        )

    def format_merge_request_description(self, *, template_meta: TemplateMeta) -> str:
        # The returned string would be treated as markdown markup.
        return self.mr_description_markdown_template.format(
            update_time=self._git_clone.update_time,
            scaraplate_version=scaraplate_version(),
            template_meta=template_meta,
        )

    def create_merge_request(self, *, title: str, description: str) -> None:
        existing_mr = self.get_merge_request()
        if existing_mr is not None:
            logger.info(f"Skipping MR creation as it already exists: {existing_mr!r}")
            return

        self._gitlab_project.mergerequests.create(
            {
                "description": description,
                "should_remove_source_branch": True,
                "source_branch": self._git_clone.changes_branch,
                "target_branch": self._gitlab_project.default_branch,
                "title": title,
            }
        )

    def get_merge_request(self):
        merge_requests = self._gitlab_project.mergerequests.list(
            state="opened",
            source_branch=self._git_clone.changes_branch,
            target_branch=self._gitlab_project.default_branch,
        )

        if not merge_requests:
            return None

        assert len(merge_requests) == 1, merge_requests
        merge_request = merge_requests[0]
        return merge_request

    @classmethod
    @contextlib.contextmanager
    def clone(
        cls,
        gitlab_url: str,
        full_project_name: str,
        private_token: str,
        *,
        mr_title_template: str = "Scheduled template update ({update_time:%Y-%m-%d})",
        mr_description_markdown_template: str = (
            "* scaraplate version: `{scaraplate_version}`\n"
            "* template commit: {template_meta.commit_url}\n"
            "* template ref: {template_meta.head_ref}\n"
        ),
        commit_author: Optional[str] = None,
        **kwargs,
    ) -> Iterator["GitLabMRProjectVCS"]:
        """Same as :meth:`.GitCloneProjectVCS.clone` with the following
        exceptions:

        - ``clone_url`` is replaced with ``gitlab_url``, ``full_project_name``
          and ``private_token``.
        - A GitLab Merge Request (MR) is opened after a successful
          ``git push``.

        The ``private_token`` allows to clone private repos, which are
        visible only for an authenticated user.

        As in :meth:`.GitCloneProjectVCS.clone`, the ``changes_branch``
        might be the same as ``clone_ref``. In this case no MR will be
        opened.

        A MR will be created only if there're any changes produced
        by scaraplate rollup. If a ``changes_branch`` is already present
        in remote (i.e. there is a previous automatic rollup which wasn't
        merged yet), there're two possibilities:

        - If one-commit diffs between the remote's ``changes_branch``
          and the local ``changes_branch`` are the same, nothing
          is done. It means that a MR already exists and it has the same
          patch as the one which was just produced locally.
        - If the diffs are different, the remote branch will be deleted,
          effectively closing the old MR, and a new one will be pushed
          instead, and a new MR will be opened.

        The opened MRs are expected to be merged manually.

        :param gitlab_url: A URL to the GitLab instance, e.g.
            ``https://gitlab.example.org``.
        :param full_project_name: Project name within gitlab, e.g.
            ``myorganization/myproject``.
        :param private_token: GitLab access token,
            see `<https://docs.gitlab.com/ce/api/#oauth2-tokens>`_.
        :param mr_title_template: :meth:`str.format` template
            which is used to produce a MR title.
            Available format variables are:

            - ``update_time`` [:class:`datetime.datetime`] -- the time
              of update
            - ``template_meta`` [:class:`.TemplateMeta`] -- template meta
              returned by :meth:`.TemplateVCS.template_meta`
        :param mr_description_markdown_template: :meth:`str.format` template
            which is used to produce a MR description (which will be rendered
            as markdown). Available format variables are:

            - ``update_time`` [:class:`datetime.datetime`] -- the time
              of update
            - ``scaraplate_version`` [:class:`str`] -- scaraplate package
              version
            - ``template_meta`` [:class:`.TemplateMeta`] -- template meta
              returned by :meth:`.TemplateVCS.template_meta`
        :param commit_author: Author name to use for ``git commit``, e.g.
            ``John Doe <john@example.org>``. If ``None``, will be retrieved
            from GitLab as the name of the currently authenticated user
            (using ``private_token``).
        """

        ensure_gitlab_is_installed()
        client = gitlab.Gitlab(url=gitlab_url, private_token=private_token, timeout=30)
        client.auth()

        gitlab_project = client.projects.get(full_project_name)
        project_url = gitlab_project_url(gitlab_url, full_project_name)
        user = client.user
        assert user is not None
        commit_author = commit_author or f"{user.name} <{user.email}>"

        # pylint wants mandatory arguments to be passed explicitly:
        changes_branch = kwargs.pop("changes_branch")

        with GitCloneProjectVCS.clone(
            clone_url=gitlab_clone_url(project_url, private_token),
            changes_branch=changes_branch,
            commit_author=commit_author,
            **kwargs,
        ) as git_clone:
            yield cls(
                git_clone,
                gitlab_project=gitlab_project,
                mr_title_template=mr_title_template,
                mr_description_markdown_template=mr_description_markdown_template,
            )



from io import BytesIO
from unittest.mock import sentinel

import pytest

from scaraplate.strategies import detect_newline


@pytest.mark.parametrize(
    "file_contents, expected_newline",
    [
        # Typical newlines:
        (BytesIO(b"a\na\n"), b"\n"),
        (BytesIO(b"a\r\na\r\n"), b"\r\n"),
        (BytesIO(b"a\ra\r"), b"\r"),
        # Mixed newlines:
        (BytesIO(b"a\r\na\na"), b"\r\n"),
        # Multiple newlines in a row:
        (BytesIO(b"a\n\n\na"), b"\n"),
        (BytesIO(b"\n\n\n"), b"\n"),
        (BytesIO(b"\r\n\r\n"), b"\r\n"),
        (BytesIO(b"\r\n"), b"\r\n"),
        # No newlines:
        (BytesIO(b"a"), sentinel.default),
        (BytesIO(b""), sentinel.default),
        (None, sentinel.default),
    ],
)
def test_detect_newline_single_file(file_contents, expected_newline):
    if file_contents is not None:
        assert file_contents.tell() == 0

    assert expected_newline == detect_newline(file_contents, default=sentinel.default)

    if file_contents is not None:
        assert file_contents.tell() == 0


@pytest.mark.parametrize(
    "file_contents1, file_contents2, expected_newline",
    [
        # One file is missing:
        (None, BytesIO(b"a\n"), b"\n"),
        (BytesIO(b"a\n"), None, b"\n"),
        # Prefer first file having a newline:
        (BytesIO(b"a\n"), BytesIO(b"a\r"), b"\n"),
        (BytesIO(b"a"), BytesIO(b"a\r"), b"\r"),
        # Fallback to default if no file contains a newline:
        (BytesIO(b"a"), BytesIO(b"a"), sentinel.default),
    ],
)
def test_detect_newline_multiple_files(
    file_contents1, file_contents2, expected_newline
):
    if file_contents1 is not None:
        assert file_contents1.tell() == 0
    if file_contents2 is not None:
        assert file_contents2.tell() == 0

    assert expected_newline == detect_newline(
        file_contents1, file_contents2, default=sentinel.default
    )

    if file_contents1 is not None:
        assert file_contents1.tell() == 0
    if file_contents2 is not None:
        assert file_contents2.tell() == 0



from unittest.mock import patch

from click.testing import CliRunner

import scaraplate.__main__ as main_module
from scaraplate.__main__ import main


def test_help():
    runner = CliRunner()
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0, result.output


def test_rollup_help():
    runner = CliRunner()
    result = runner.invoke(main, ["rollup", "--help"])
    assert result.exit_code == 0, result.output


def test_extra_context():
    with patch.object(main_module, "_rollup") as mock_rollup:
        runner = CliRunner()
        result = runner.invoke(
            main, ["rollup", ".", "mydest", "key1=value1", "key2=value2"]
        )
        assert result.exit_code == 0, result.output

        assert mock_rollup.call_count == 1
        _, kwargs = mock_rollup.call_args
        assert kwargs["extra_context"] == {"key1": "value1", "key2": "value2"}


def test_extra_context_incorrect():
    with patch.object(main_module, "_rollup"):
        runner = CliRunner()
        result = runner.invoke(main, ["rollup", ".", "mydest", "key1value1"])
        assert result.exit_code == 2, result.output
        assert (
            "EXTRA_CONTEXT should contain items of the form key=value" in result.output
        )



import pytest

from scaraplate.template import (
    _git_head_commit_hash,
    _git_remote_origin,
    _git_resolve_head,
    _is_git_dirty,
)


def test_git_head_commit_hash_valid(init_git_and_commit, tempdir_path):
    repo_path = tempdir_path / "repo"
    repo_path.mkdir()
    (repo_path / "myfile").write_text("hi")
    init_git_and_commit(repo_path)

    assert 40 == len(_git_head_commit_hash(repo_path))


def test_git_head_commit_hash_invalid(tempdir_path):
    # tempdir_path is not under a git repo.
    with pytest.raises(RuntimeError):
        _git_head_commit_hash(tempdir_path)


def test_git_head_ref_valid(init_git_and_commit, tempdir_path):
    repo_path = tempdir_path / "repo"
    repo_path.mkdir()
    (repo_path / "myfile").write_text("hi")
    init_git_and_commit(repo_path)

    assert "master" == _git_resolve_head(repo_path)


def test_git_head_ref_valid_detached(init_git_and_commit, call_git, tempdir_path):
    repo_path = tempdir_path / "repo"
    repo_path.mkdir()
    (repo_path / "myfile").write_text("hi")
    init_git_and_commit(repo_path)

    # Detach the HEAD
    commit_hash = _git_head_commit_hash(repo_path)
    call_git(f"git checkout {commit_hash}", cwd=repo_path)

    assert _git_resolve_head(repo_path) is None


def test_git_head_ref_invalid(tempdir_path):
    # tempdir_path is not under a git repo.
    with pytest.raises(RuntimeError):
        _git_resolve_head(tempdir_path)


def test_git_is_dirty(init_git_and_commit, tempdir_path, call_git):
    repo_path = tempdir_path / "repo"
    repo_path.mkdir()
    (repo_path / "myfile").write_text("hi")
    init_git_and_commit(repo_path)

    assert not _is_git_dirty(repo_path)

    # Unstaged should be considered dirty
    (repo_path / "myfile2").write_text("hi")
    assert _is_git_dirty(repo_path)

    # Staged should be considered dirty as well
    call_git("git add myfile2", cwd=repo_path)
    assert _is_git_dirty(repo_path)


def test_git_remote_origin_valid(init_git_and_commit, tempdir_path):
    repo_path = tempdir_path / "repo"
    repo_path.mkdir()
    (repo_path / "myfile").write_text("hi")
    init_git_and_commit(repo_path)

    assert _git_remote_origin(repo_path).startswith("https://")


def test_git_remote_origin_invalid(init_git_and_commit, tempdir_path):
    repo_path = tempdir_path / "repo"
    repo_path.mkdir()
    (repo_path / "myfile").write_text("hi")
    init_git_and_commit(repo_path, with_remote=False)

    with pytest.raises(RuntimeError):
        _git_remote_origin(repo_path)



from pathlib import Path

import pytest

from scaraplate.automation.git import Git


@pytest.fixture
def git_repo(init_git_and_commit, tempdir_path):
    repo_path = tempdir_path / "repo"
    repo_path.mkdir()
    (repo_path / "readme").write_text("hi")
    init_git_and_commit(repo_path, with_remote=False)

    return repo_path


@pytest.mark.parametrize(
    "remote, ref, expected_ref", [("origin", "my-branch", "origin/my-branch")]
)
def test_git_remote_ref(remote, ref, expected_ref):
    git = Git(cwd=Path("."), remote=remote)
    assert expected_ref == git.remote_ref(ref)


def test_git_commit(git_repo, call_git):
    git = Git(git_repo)
    assert not git.is_dirty()

    (git_repo / "myfile").write_text("my precious")
    assert git.is_dirty()

    message = "my\nmultiline\nmessage"
    author = "John Doe <whitehouse@gov>"
    git.commit_all(message, author=author)
    assert not git.is_dirty()

    assert message == call_git("git log -1 --pretty=%B", cwd=git_repo)
    assert author == call_git("git log -1 --pretty='%an <%ae>'", cwd=git_repo)


def test_git_is_existing_ref(git_repo):
    git = Git(git_repo)
    assert git.is_existing_ref("master")
    assert not git.is_existing_ref("origin/master")


def test_git_are_one_commit_diffs_equal(git_repo, call_git):
    git = Git(git_repo)

    with pytest.raises(RuntimeError):
        # These refs don't exist yet
        git.are_one_commit_diffs_equal("t1", "t2")

    # Create branch t1
    git.checkout_branch("t1")
    (git_repo / "common").write_text("same")
    git.commit_all("m1")

    # Move master 1 commit ahead
    call_git("git checkout master", cwd=git_repo)
    (git_repo / "some_file").write_text("i'ma sta")
    call_git("git add .", cwd=git_repo)
    call_git("git commit -m next", cwd=git_repo)

    # Create branch t2
    git.checkout_branch("t2")
    (git_repo / "common").write_text("same")
    git.commit_all("m2")

    #      /-()  << t1 (/common == "same")
    #     /
    # -> () -> () -> ()  << t2 (/common == "same")
    #           \
    #            \  << master (/some_file == "i'ma sta")
    assert git.are_one_commit_diffs_equal("t1", "t2")
    assert git.are_one_commit_diffs_equal("master", "master")
    assert not git.are_one_commit_diffs_equal("master", "t2")



import json
from unittest.mock import call, patch, sentinel

import pytest

import scaraplate.automation.gitlab
from scaraplate.automation.gitlab import (
    GitLabCloneTemplateVCS,
    GitLabMRProjectVCS,
    gitlab_available,
    gitlab_clone_url,
    gitlab_project_url,
)
from scaraplate.template import TemplateMeta


@pytest.mark.parametrize(
    "project_url, private_token, expected_url",
    [
        (
            "https://mygitlab.com/myorg/myproj.git",
            "xxxtokenxxx",
            "https://oauth2:xxxtokenxxx@mygitlab.com/myorg/myproj.git",
        ),
        (
            "https://mygitlab.com/myorg/myproj",
            "xxxtokenxxx",
            "https://oauth2:xxxtokenxxx@mygitlab.com/myorg/myproj.git",
        ),
        (
            "https://mygitlab.com/myorg/myproj",
            None,
            "https://mygitlab.com/myorg/myproj.git",
        ),
    ],
)
def test_gitlab_clone_url(project_url, private_token, expected_url):
    assert expected_url == gitlab_clone_url(project_url, private_token)


@pytest.mark.parametrize(
    "gitlab_url, full_project_name, expected_url",
    [
        ("https://mygitlab.com", "myorg/myproj", "https://mygitlab.com/myorg/myproj"),
        (
            "https://mygitlab.com/",
            "myorg/myproj.git",
            "https://mygitlab.com/myorg/myproj.git",
        ),
    ],
)
def test_gitlab_project_url(gitlab_url, full_project_name, expected_url):
    assert expected_url == gitlab_project_url(gitlab_url, full_project_name)


def test_gitlab_clone_template(template_bare_git_repo, call_git):
    with patch.object(
        scaraplate.automation.gitlab,
        "gitlab_clone_url",
        return_value=str(template_bare_git_repo),
    ) as mock_gitlab_clone_url:
        with GitLabCloneTemplateVCS.clone(
            project_url=sentinel.project_url, private_token=sentinel.private_token
        ) as template_vcs:
            # Ensure this is a valid git repo:
            call_git("git status", cwd=template_vcs.dest_path)

            assert template_vcs.template_meta.head_ref == "master"
        assert mock_gitlab_clone_url.call_count == 1
        assert mock_gitlab_clone_url.call_args == call(
            sentinel.project_url, sentinel.private_token
        )


@pytest.mark.skipif(
    gitlab_available, reason="gitlab should not be installed for this test"
)
def test_gitlab_mr_project_raises_without_gitlab():
    with pytest.raises(ImportError):
        with GitLabMRProjectVCS.clone(
            gitlab_url="http://a",
            full_project_name="a/aa",
            private_token="zzz",
            changes_branch="update",
        ):
            raise AssertionError("should not be executed")


@pytest.mark.skipif(
    not gitlab_available, reason="gitlab should be installed for this test"
)
@pytest.mark.parametrize(
    "mr_exists, changes_branch",
    [
        (False, "update"),
        (True, "update"),
        (None, "master"),  # MRs cannot be created from default branch
    ],
)
def test_gitlab_mr_project(
    project_bare_git_repo, call_git, http_server, mr_exists, changes_branch
):
    requests = iter(
        [
            (
                "GET",
                "/api/v4/user",
                {
                    # https://docs.gitlab.com/ee/api/users.html#list-current-user-for-normal-users
                    "name": "test user",
                    "email": "test@notemail",
                },
                None,
            ),
            (
                "GET",
                "/api/v4/projects/a%2Faa",
                {
                    # https://docs.gitlab.com/ee/api/projects.html#get-single-project
                    "id": 42,
                    "name": "test project",
                    "default_branch": "master",
                },
                None,
            ),
        ]
        + (
            []
            if mr_exists is None
            else [
                (
                    "GET",
                    "/api/v4/projects/42/merge_requests"
                    "?state=opened&source_branch=update&target_branch=master",
                    [] if not mr_exists else [{"id": 98, "iid": 98}],
                    None,
                )
            ]
        )
        + (
            []
            if mr_exists or mr_exists is None
            else [
                (
                    "POST",
                    "/api/v4/projects/42/merge_requests",
                    {"id": 99, "iid": 99},
                    lambda body: mr_body_checker(body),
                )
            ]
        )
    )

    def mr_body_checker(body):
        data = json.loads(body.decode())
        assert "1111111111111111111111111111111111111111" in data["description"]

    @http_server.set_request_handler
    def request_handler(method, path, headers, body):
        assert headers["PRIVATE-TOKEN"] == "zzz"
        try:
            exp_method, exp_path, response, body_checker = next(requests)
        except StopIteration:  # pragma: no cover
            raise AssertionError(f"Received unexpected request {method} {path}")
        assert method == exp_method
        assert path == exp_path
        if body_checker is not None:
            body_checker(body)
        return 200, {"Content-type": "application/json"}, json.dumps(response).encode()

    with patch.object(
        scaraplate.automation.gitlab,
        "gitlab_clone_url",
        return_value=str(project_bare_git_repo),
    ) as mock_gitlab_clone_url:
        with GitLabMRProjectVCS.clone(
            gitlab_url=http_server.get_url(),
            full_project_name="a/aa",
            private_token="zzz",
            changes_branch=changes_branch,
        ) as project_vcs:
            # Ensure this is a valid git repo:
            call_git("git status", cwd=project_vcs.dest_path)

            assert not project_vcs.is_dirty()

            (project_vcs.dest_path / "hi").write_text("change!")
            project_vcs.commit_changes(
                TemplateMeta(
                    git_project_url="https://testdomain/t/tt",
                    commit_hash="1111111111111111111111111111111111111111",
                    commit_url=(
                        "https://testdomain/t/tt"
                        "/commit/1111111111111111111111111111111111111111"
                    ),
                    is_git_dirty=False,
                    head_ref="master",
                )
            )

            commit_message = call_git(
                f'git log --pretty=format:"%B" -n 1 {changes_branch}',
                cwd=project_bare_git_repo,
            )
            assert "1111111111111111111111111111111111111111" in commit_message
            assert "Scheduled" in commit_message

            commit_author = call_git(
                f'git log --pretty=format:"%aN <%aE>" -n 1 {changes_branch}',
                cwd=project_bare_git_repo,
            )
            assert commit_author == "test user <test@notemail>"

            assert "change!" == call_git(
                f"git show {changes_branch}:hi", cwd=project_bare_git_repo
            )

        assert mock_gitlab_clone_url.call_count == 1
        assert mock_gitlab_clone_url.call_args == call(
            f"{http_server.get_url()}/a/aa", "zzz"
        )

        # ensure that all expected queries have been executed:
        assert next(requests, ...) is ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
import os
import subprocess

from setuptools import setup
from setuptools.command.sdist import sdist as sdist_orig

ROOT = os.path.dirname(__file__)
VERSION = os.path.join(ROOT, "VERSION")


def main():
    return setup(
        cmdclass={
            "sdist": sdist,
        },
        version=project_version(),
    )


def project_version():
    version = None

    if not version:
        try:
            output = (
                subprocess.check_output(
                    ["git", "describe", "--tags", "--always"],
                    stderr=open(os.devnull, "wb"),
                )
                .strip()
                .decode()
            )
        except (FileNotFoundError, subprocess.CalledProcessError):
            # Can't read the tag. That's probably project at initial stage,
            # or git is not available at all or something else.
            pass
        else:
            try:
                base, distance, commit_hash = output.split("-")
            except ValueError:
                # We're on release tag.
                version = output
            else:
                # Reformat git describe for PEP-440
                version = "{}.{}+{}".format(base, distance, commit_hash)

    if not version and os.path.exists(VERSION):
        with open(VERSION) as verfile:
            version = verfile.read().strip()

    if not version:
        raise RuntimeError("cannot detect project version")

    return version


class sdist(sdist_orig):
    def run(self):
        # In case when user didn't eventually run `make version` ensure that
        # VERSION file will be included in source distribution.
        version = project_version()
        with open(VERSION, "w") as verfile:
            verfile.write(version)
        sdist_orig.run(self)


if __name__ == "__main__":
    main()



import contextlib
import fnmatch
import io
import os
import pprint
import tempfile
from pathlib import Path
from typing import BinaryIO, Mapping, Optional, Tuple, Union

import click
from cookiecutter.main import cookiecutter

from .config import (
    ScaraplateYamlOptions,
    ScaraplateYamlStrategies,
    StrategyNode,
    get_scaraplate_yaml_options,
    get_scaraplate_yaml_strategies,
)
from .cookiecutter import CookieCutterContextDict
from .template import TemplateMeta, get_template_meta_from_git

__all__ = ("rollup", "InvalidScaraplateTemplateError")


class InvalidScaraplateTemplateError(Exception):
    """Something is wrong with the template.

    .. versionadded:: 0.2
    """


def rollup(
    template_dir: Union[Path, str],
    target_project_dir: Union[Path, str],
    *,
    no_input: bool,
    extra_context: Optional[Mapping[str, str]] = None,
) -> None:
    """The essence of scaraplate: the rollup function, which can be called
    from Python.

    Raises :class:`.InvalidScaraplateTemplateError`.

    .. versionchanged:: 0.2
        Added `extra_context` argument which supersedes
        :class:`scaraplate.cookiecutter.CookieCutterContext` context.
        Useful to provide context values non-interactively.

        ``scaraplate.yaml`` now supports cookiecutter variables
        in the `strategies_mapping`` keys.
    """
    template_path = Path(template_dir)
    target_path = Path(target_project_dir)

    scaraplate_yaml_options = get_scaraplate_yaml_options(template_path)
    template_meta = get_template_meta_from_git(
        template_path, git_remote_type=scaraplate_yaml_options.git_remote_type
    )

    target_path.mkdir(parents=True, exist_ok=True, mode=0o755)
    project_dest = get_project_dest(target_path)

    cookiecutter_context = get_target_project_cookiecutter_context(
        target_path, scaraplate_yaml_options
    )
    cookiecutter_context = CookieCutterContextDict(
        {**cookiecutter_context, **(extra_context or {})}
    )

    with tempfile.TemporaryDirectory() as tempdir_path:
        output_dir = Path(tempdir_path) / "out"
        output_dir.mkdir(mode=0o700)
        if not no_input:
            click.echo(f'`project_dest` must equal to "{project_dest}"')

        # By default cookiecutter preserves all entered variables in
        # the user's home and tries to reuse it on subsequent cookiecutter
        # executions.
        #
        # Give cookiecutter a fake config so it would write its stuff
        # to the tempdir, which would then be removed.
        cookiecutter_config_path = Path(tempdir_path) / "cookiecutter_home"
        cookiecutter_config_path.mkdir(parents=True)
        cookiecutter_config = cookiecutter_config_path / "cookiecutterrc.yaml"
        cookiecutter_config.write_text(
            f"""
cookiecutters_dir: "{cookiecutter_config_path / 'cookiecutters'}"
replay_dir: "{cookiecutter_config_path / 'replay'}"
"""
        )

        cookiecutter_context.setdefault(  # pylint: disable=no-member
            "project_dest", project_dest
        )

        template_root_path, template_dir_name = get_template_root_and_dir(template_path)

        with with_cwd(template_root_path):
            # Cookiecutter preserves its template values to
            # .scaraplate.conf/setup.cfg (this is specified in the template).
            #
            # These values contain a `_template` key, which points to
            # the template just like it was passed to cookiecutter
            # (that is the first positional arg in the command below).
            #
            # In order to specify only the directory name of the template,
            # we change cwd (current working directory) and pass
            # the path to template as just a directory name, effectively
            # stripping off the path to the template in the local
            # filesystem.
            cookiecutter(
                template_dir_name,
                no_input=no_input,
                extra_context=cookiecutter_context,
                output_dir=str(output_dir.resolve()),
                config_file=str(cookiecutter_config),
            )

        # Say the `target_path` looks like `/some/path/to/myproject`.
        #
        # Cookiecutter puts the generated project files to
        # `output_path / "{{ cookiecutter.project_dest }}"`.
        #
        # We assume that `project_dest` equals the dirname of
        # the `target_path` (i.e. `myproject`).
        #
        # Later we need to move the generated files from
        # `output_path / "{{ cookiecutter.project_dest }}" / ...`
        # to `/some/path/to/myproject/...`.
        #
        # For that we need to ensure that cookiecutter did indeed
        # put the generated files to `output_path / 'myproject'`
        # and no extraneous files or dirs were generated.
        actual_items_in_tempdir = os.listdir(output_dir)
        expected_items_in_tempdir = [project_dest]
        if actual_items_in_tempdir != expected_items_in_tempdir:
            raise RuntimeError(
                f"A project generated by cookiecutter has an unexpected "
                f"file structure.\n"
                f"Expected directory listing: {expected_items_in_tempdir}\n"
                f"Actual: {actual_items_in_tempdir}\n"
                f"\n"
                f"Does the TARGET_PROJECT_DIR name match "
                f"the cookiecutter's `project_dest` value?"
            )

        with with_cwd(output_dir / project_dest):
            # Pass a relative path to CookieCutterContext so the __str__
            # wouldn't include a full absolute path to the temp dir.
            cookiecutter_context_dict = get_cookiecutter_context_from_temp_project(
                Path("."), scaraplate_yaml_options
            )

        scaraplate_yaml_strategies = get_scaraplate_yaml_strategies(
            template_path, cookiecutter_context_dict
        )
        apply_generated_project(
            output_dir / project_dest,
            target_path,
            template_meta=template_meta,
            scaraplate_yaml_strategies=scaraplate_yaml_strategies,
        )

        click.echo("Done!")


def get_project_dest(target_dir_path: Path) -> str:
    return target_dir_path.resolve().name


def get_template_root_and_dir(template_path: Path) -> Tuple[Path, str]:
    template_resolved_path = template_path.resolve()
    template_root_path = template_resolved_path.parents[0]
    template_dir_name = template_resolved_path.name
    return template_root_path, template_dir_name


def get_target_project_cookiecutter_context(
    target_path: Path, scaraplate_yaml_options: ScaraplateYamlOptions
) -> CookieCutterContextDict:
    cookiecutter_context = scaraplate_yaml_options.cookiecutter_context_type(
        target_path
    )

    try:
        context: CookieCutterContextDict = cookiecutter_context.read()
    except FileNotFoundError:
        click.echo(
            f"`{cookiecutter_context}` file doesn't exist, "
            f"continuing with an empty context..."
        )
        return CookieCutterContextDict({})
    else:
        if context:
            click.echo(
                f"Continuing with the following context from "
                f"the `{cookiecutter_context}` file:\n{pprint.pformat(context)}"
            )
        else:
            click.echo(
                f"No context found in the `{cookiecutter_context}` file, "
                f"continuing with an empty one..."
            )
        return CookieCutterContextDict(dict(context))


def get_cookiecutter_context_from_temp_project(
    target_path: Path, scaraplate_yaml_options: ScaraplateYamlOptions
) -> CookieCutterContextDict:
    cookiecutter_context = scaraplate_yaml_options.cookiecutter_context_type(
        target_path
    )

    try:
        context: CookieCutterContextDict = cookiecutter_context.read()
    except FileNotFoundError:
        raise InvalidScaraplateTemplateError(
            f"cookiecutter context file `{cookiecutter_context}` doesn't exist "
            f"in the rendered template. Ensure you have added its generation "
            f"to the template. See docs for {type(cookiecutter_context)}"
        )
    else:
        if not context:
            raise InvalidScaraplateTemplateError(
                f"cookiecutter context file `{cookiecutter_context}` is empty "
                f"in the rendered template. Ensure you have correctly added its "
                f"generation to the template. See docs for {type(cookiecutter_context)}"
            )
    return CookieCutterContextDict(dict(context))


def apply_generated_project(
    generated_path: Path,
    target_path: Path,
    *,
    template_meta: TemplateMeta,
    scaraplate_yaml_strategies: ScaraplateYamlStrategies,
) -> None:
    generated_path = generated_path.resolve()

    for root, dirs, files in os.walk(generated_path):
        current_root_path = Path(root)
        path_from_template_root = current_root_path.relative_to(generated_path)
        target_root_path = target_path / path_from_template_root
        target_root_path.mkdir(parents=True, exist_ok=True, mode=0o755)

        for d in dirs:
            (target_root_path / d).mkdir(parents=True, exist_ok=True, mode=0o755)

        for f in files:
            file_path = current_root_path / f
            target_file_path = target_root_path / f

            strategy_node = get_strategy(
                scaraplate_yaml_strategies, path_from_template_root / f
            )

            template_contents = io.BytesIO(file_path.read_bytes())
            if target_file_path.exists():
                target_contents: Optional[BinaryIO] = io.BytesIO(
                    target_file_path.read_bytes()
                )
            else:
                target_contents = None

            strategy = strategy_node.strategy(
                target_contents=target_contents,
                template_contents=template_contents,
                template_meta=template_meta,
                config=strategy_node.config,
            )

            target_contents = strategy.apply()
            target_file_path.write_bytes(target_contents.read())

            # https://stackoverflow.com/a/5337329
            chmod = file_path.stat().st_mode & 0o777
            target_file_path.chmod(chmod)


def get_strategy(
    scaraplate_yaml_strategies: ScaraplateYamlStrategies, path: Path
) -> StrategyNode:
    for glob_pattern, strategy_node in sorted(
        scaraplate_yaml_strategies.strategies_mapping.items()
    ):
        if fnmatch.fnmatch(str(path), glob_pattern):
            return strategy_node
    return scaraplate_yaml_strategies.default_strategy


@contextlib.contextmanager
def with_cwd(cwd: Path):
    initial_cwd = os.getcwd()
    try:
        os.chdir(cwd)
        yield
    finally:
        os.chdir(initial_cwd)
